""" Contains the RowGenerator subclass for generating rows from a CompLib composition. """

import json
from typing import Dict, Optional, List, Tuple
from urllib.parse import urlparse

import requests

from wheatley.aliases import Row
from wheatley.stroke import Stroke
from wheatley.bell import Bell
from .row_generator import RowGenerator


class PrivateCompError(ValueError):
    """ An error class thrown when a user tries to access a private CompLib composition. """

    def __init__(self, comp_id: int) -> None:
        super().__init__()

        self._comp_id = comp_id

    def __str__(self) -> str:
        return f"Composition with ID {self._comp_id} is private."


class InvalidCompError(ValueError):
    """ An error class thrown when a user tries to access an invalid CompLib composition. """

    def __init__(self, comp_id: int) -> None:
        super().__init__()

        self._comp_id = comp_id

    def __str__(self) -> str:
        return f"Composition with ID {self._comp_id} is does not exist."


class InvalidComplibURLError(Exception):
    """ An error class thrown when a user inputs an invalid complib comp URL. """

    def __init__(self, url: str, error: str) -> None:
        super().__init__()

        self._url = url
        self._error = error

    def __str__(self) -> str:
        return f"Invalid CompLib URL '{self._url}': {self._error}"


# The 'removeprefix' function from Python 3.9
def removeprefix(string: str, prefix: str) -> str:
    """ Removes a prefix from a given string if it exists. """
    if string.startswith(prefix):
        return string[len(prefix):]
    return string[:]


def parse_arg(arg: str) -> Tuple[int, Optional[str], Optional[int]]:
    """ Parses a URL or a plain ID into a tuple of (ID, access key, substituted method ID). """
    # === CONVERT ID NUMBER INTO A NORMALISED URL ===
    # If the arg doesn't contain 'complib.org', we assume it's simply an ID (perhaps with an access
    # key) and so turn it into a URL
    if "complib.org" in arg:
        url = arg
    else:
        url = "https://complib.org/composition/" + arg
    # Make sure that the URL starts with `https://`
    if not url.startswith('http'):
        url = 'https://' + url

    # === PARSE URL, AND SPLIT PATH SEGMENTS ===
    # `url` looks like a URL, so try to parse it
    parsed_url = urlparse(url)
    # Naively split the URL path with `/`s.  This is technically wrong because it doesn't handle
    # escaping properly, but we don't worry about that because CompLib URLs shouldn't have any
    # escaped `/`s anyway.
    path_segs = parsed_url.path.split("/")

    # === PARSE COMP ID FROM PATH SEGMENTS ===
    # Check that we have an absolute path, and remove the '' at the front of the list of path
    # segments
    assert path_segs[0] == ''
    path_segs = path_segs[1:]
    # Raise helpful error messages if obvious things are wrong (these errors can't happen if the
    # user used a straight ID)
    if len(path_segs) <= 1:
        raise InvalidComplibURLError(url, "URL needs more path segments.")
    if path_segs[0] != "composition":
        raise InvalidComplibURLError(url, "Not a composition.")
    # Parse the 2nd URL segment into the comp ID
    try:
        comp_id = int(path_segs[1])
    except ValueError as e:
        raise InvalidComplibURLError(url, f"Composition ID '{path_segs[1]}' is not a number.") from e

    # === PARSE ACCESS KEY FROM THE QUERY STRING ===
    access_key = None
    substituted_method_id = None
    for q in parsed_url.query.split("&"):
        parts = q.split("=")
        if len(parts) != 2:
            continue
        if parts[0] == "accessKey":
            access_key = parts[1]
        if parts[0] == "substitutedmethodid":
            try:
                substituted_method_id = int(parts[1])
            except ValueError as e:
                raise InvalidComplibURLError(
                    url,
                    f"Substituted method ID '{parts[1]}' is not a number."
                ) from e

    return (comp_id, access_key, substituted_method_id)


class ComplibCompositionGenerator(RowGenerator):
    """ The RowGenerator subclass for generating rows from a CompLib composition. """

    complib_url = "https://api.complib.org/composition/"

    def __init__(self, comp_id: int, access_key: Optional[str] = None,
                 substituted_method_id: Optional[int] = None) -> None:
        def process_call_string(calls: str) -> List[str]:
            """ Parse a sequence of calls, and remove 'Stand'. """
            stripped_calls = [x.strip() for x in calls.split(";")]
            return [c for c in stripped_calls if c != 'Stand']

        # Generate URL from the params
        query_sections: List[str] = []
        if access_key:
            query_sections.append(f"accessKey={access_key}")
        if substituted_method_id:
            query_sections.append(f"substitutedmethodid={substituted_method_id}")
        url = self.complib_url + str(comp_id) + "/rows"
        if query_sections:
            url += "?" + "&".join(query_sections)
        # Create an HTTP request for the rows, and deal with potential error codes
        request_rows = requests.get(url)
        if request_rows.status_code == 404:
            raise InvalidCompError(comp_id)
        if request_rows.status_code == 403:
            raise PrivateCompError(comp_id)
        request_rows.raise_for_status()
        # Parse the request responses as JSON
        response_rows = json.loads(request_rows.text)
        unparsed_rows = response_rows['rows']
        # Determine the start row of the composition
        num_starting_rounds = 0
        while unparsed_rows[num_starting_rounds][0] == unparsed_rows[0][0]:
            num_starting_rounds += 1
        self._start_stroke = Stroke.from_index(num_starting_rounds)
        # Derive the rows, calls and stage from the JSON response
        loaded_rows: List[Tuple[Row, List[str]]] = [
            (
                Row([Bell.from_str(bell) for bell in row]),
                [] if calls == '' else process_call_string(calls)
            )
            for row, calls, _property_bitmap in unparsed_rows
        ]
        # Convert these parsed rows into a format that we can read more easily when ringing.  I.e.
        # load the non-rounds rows as-is, but keep the calls that should be called in rounds
        self.loaded_rows = loaded_rows[num_starting_rounds:]
        self._early_calls = {
            num_starting_rounds - i: calls
            for i, (_row, calls) in enumerate(loaded_rows[:num_starting_rounds])
            if calls != []
        }
        # Set the variables from which the summary string is generated
        self.comp_id = comp_id
        self.comp_title = response_rows['title']
        self.is_comp_private = access_key is not None
        # Propogate the initialisation up to the parent class
        super().__init__(response_rows['stage'])

    @classmethod
    def from_arg(cls, arg: str) -> 'ComplibCompositionGenerator':
        """ Generates a ComplibCompositionGenerator from either a URL or the ID of a comp. """
        comp_id, access_key, substituted_method_id = parse_arg(arg)
        return cls(comp_id, access_key, substituted_method_id)

    def summary_string(self) -> str:
        """ Returns a short string summarising the RowGenerator. """
        return f"{'private ' if self.is_comp_private else ''}comp #{self.comp_id}: {self.comp_title}"

    def _gen_row_and_calls(self, _previous_row: Row, _stroke: Stroke, index: int) -> Tuple[Row, List[str]]:
        if index < len(self.loaded_rows):
            return self.loaded_rows[index]
        return (self.rounds(), [])

    def _gen_row(self, previous_row: Row, stroke: Stroke, index: int) -> Row:
        # Technically, this should be unreachable, but Python doesn't have an error for that (I miss
        # Rust's `unreachable!` macro)
        raise NotImplementedError()

    def start_stroke(self) -> Stroke:
        """ Gets the stroke of the first row.  We allow backstroke starts, and derive this in the constructor
        """
        return self._start_stroke

    def early_calls(self) -> Dict[int, List[str]]:
        """ See row_generator/row_generator.py for documentation. """
        return self._early_calls



"""
A module to store functions related to parsing the HTML of the Ringing Room tower pages to find
things like the load-balanced URL of the socket-io server.
"""

import urllib
import requests


class TowerNotFoundError(ValueError):
    """ An error class created whenever the user inputs an incorrect room id. """

    def __init__(self, tower_id: int, url: str) -> None:
        super().__init__()

        self._id = tower_id
        self._url = url

    def __str__(self) -> str:
        return f"Tower {self._id} not found at '{self._url}'."

class InvalidURLError(Exception):
    """ An error class created whenever the user inputs a URL that is invalid. """

    def __init__(self, url: str) -> None:
        super().__init__()

        self._url = url

    def __str__(self) -> str:
        return f"Unable to make a connection to '{self._url}'."


def _fix_url(url: str) -> str:
    """ Add 'https://' to the start of a URL if necessary """
    corrected_url = url if url.startswith("http") else "https://" + url

    return corrected_url


def get_load_balancing_url(tower_id: int, unfixed_http_server_url: str) -> str:
    """
    Get the URL of the socket server which (since the addition of load balancing) is not
    necessarily the same as the URL of the http server that people will put into their browser URL
    bars.
    """
    http_server_url = _fix_url(unfixed_http_server_url)
    url = urllib.parse.urljoin(http_server_url, str(tower_id)) # type: ignore

    try:
        html = requests.get(url).text
    except requests.exceptions.ConnectionError as e:
        raise InvalidURLError(http_server_url) from e

    try:
        # Trying to extract the following line in the rendered html:
        # server_ip: "{{server_ip}}"
        # See https://github.com/lelandpaul/virtual-ringing-room/blob/
        #     ec00927ca57ab94fa2ff6a978ffaff707ab23a57/app/templates/ringing_room.html#L46
        url_start_index = html.index("server_ip") + len('server_ip: "')
        string_that_starts_with_url = html[url_start_index:]
        load_balancing_url = string_that_starts_with_url[:string_that_starts_with_url.index('"')]

        return load_balancing_url
    except ValueError as e:
        raise TowerNotFoundError(tower_id, http_server_url) from e



"""
Module to contain the Bot class, which acts as 'glue' to combine the rhythm, row generation and
socket-io communication provided by the Rhythm, RowGenerator and Tower objects into a useful
program.
"""

import time
import logging
from typing import Optional, Any, List

from wheatley import calls
from wheatley.aliases import JSON, Row
from wheatley.stroke import Stroke
from wheatley.bell import Bell
from wheatley.rhythm import Rhythm
from wheatley.row_generation.helpers import generate_starting_row, rounds
from wheatley.tower import RingingRoomTower
from wheatley.parsing import to_bool, json_to_row_generator, RowGenParseError
from wheatley.row_generation import RowGenerator


# Number of seconds that Wheatley is not ringing before Wheatley will return from the mainloop
# Only applies when Wheatley is running in server mode
INACTIVITY_EXIT_TIME = 300


# Bot holds a lot of state, allow it to have more fields
# pylint: disable=too-many-instance-attributes
class Bot:
    """
    A class to hold all the information that Wheatley will use to glue together the rhythm,
    row_gen and socket-io parts together into a useful program.
    """

    logger_name = "BOT"

    def __init__(self, tower: RingingRoomTower, row_generator: RowGenerator, do_up_down_in: bool,
                 stop_at_rounds: bool, call_comps: bool, rhythm: Rhythm, user_name: Optional[str] = None,
                 server_instance_id: Optional[int] = None) -> None:
        """ Initialise a Bot with all the parts it needs to run. """
        # If this is None then Wheatley is in client mode, otherwise Wheatley is in server mode
        self._server_instance_id = server_instance_id
        self._last_activity_time = time.time()

        self._rhythm = rhythm
        self._do_up_down_in = do_up_down_in
        self._stop_at_rounds = stop_at_rounds
        self._call_comps = call_comps
        self._user_name = user_name

        self.row_generator = row_generator
        # This is the row generator that will be used after 'Look to' is called for the next time,
        # allowing for changing the method or composition whilst Wheatley is running.
        self.next_row_generator: Optional[RowGenerator] = None

        self._tower = tower

        self._tower.invoke_on_call[calls.LOOK_TO].append(self._on_look_to)
        self._tower.invoke_on_call[calls.GO].append(self._on_go)
        self._tower.invoke_on_call[calls.BOB].append(self._on_bob)
        self._tower.invoke_on_call[calls.SINGLE].append(self._on_single)
        self._tower.invoke_on_call[calls.THATS_ALL].append(self._on_thats_all)
        self._tower.invoke_on_call[calls.STAND].append(self._on_stand_next)
        self._tower.invoke_on_bell_rung.append(self._on_bell_ring)
        self._tower.invoke_on_reset.append(self._on_size_change)
        if self._server_mode:
            self._tower.invoke_on_setting_change.append(self._on_setting_change)
            self._tower.invoke_on_row_gen_change.append(self._on_row_gen_change)
            self._tower.invoke_on_stop_touch.append(self._on_stop_touch)

        self._is_ringing = False
        self._is_ringing_rounds = True
        self._is_ringing_opening_row = True
        # This is used as a counter - once `Go` or `Look To` is received, the number of rounds left
        # is calculated and then decremented at the start of every subsequent row until it reaches
        # 0, at which point the method starts.  We keep a counter rather than a simple flag so that
        # calls can be called **before** going into changes when Wheatley is calling (useful for
        # calling the first method name in spliced and early calls in Original, Erin, etc.).  The
        # value `None` is used to represent the case where we don't know when we will be starting
        # the method (and therefore there it makes no sense to decrement this counter).
        self._rounds_left_before_method: Optional[int] = None
        self._should_start_ringing_rounds = False
        self._should_stand = False

        self._row_number = 0
        self._place = 0
        self._opening_row: Row = row_generator.start_row
        self._rounds: Row = rounds(self.number_of_bells)
        self._row: Row = self._rounds

        # This is used because the row's calls are generated at the **end** of each row (or on
        # `Look To`), but need to be called at the **start** of the next row.
        self._calls: List[str] = []

        self.logger = logging.getLogger(self.logger_name)

        # Log what we're going to ring, and how to stop Wheatley
        self.logger.info(f"Wheatley will ring {self.row_generator.summary_string()}")
        self.logger.info("Press `Control-C` to stop Wheatley ringing, e.g. to change method.")

    # Convenient properties that are frequently used
    @property
    def stroke(self) -> Stroke:
        """ Returns true if the current row (determined by self._row_number) represents a handstroke. """
        return Stroke.from_index(self._row_number)

    @property
    def number_of_bells(self) -> int:
        """ Convenient property to find the number of bells in the current tower. """
        return self._tower.number_of_bells

    @property
    def _server_mode(self) -> bool:
        return self._server_instance_id is not None

    # Callbacks
    def _on_setting_change(self, key: str, value: Any) -> None:
        def log_invalid_key(message: str) -> None:
            self.logger.warning(f"Invalid value for {key}: {message}")

        if key == 'use_up_down_in':
            try:
                self._do_up_down_in = to_bool(value)
                self.logger.info(f"Setting 'use_up_down_in' to {self._do_up_down_in}")
            except ValueError:
                log_invalid_key(f"{value} cannot be converted into a bool")
        elif key == 'stop_at_rounds':
            try:
                self._stop_at_rounds = to_bool(value)
                self.logger.info(f"Setting 'stop_at_rounds' to {value}")
            except ValueError:
                log_invalid_key(f"{value} cannot be converted into a bool")
        else:
            self._rhythm.change_setting(key, value, time.time())

    def _on_row_gen_change(self, row_gen_json: JSON) -> None:
        try:
            self.next_row_generator = json_to_row_generator(row_gen_json, self.logger)
            self.logger.info(f"Next touch, Wheatley will ring {self.next_row_generator.summary_string()}")
        except RowGenParseError as e:
            self.logger.warning(e)

    def _on_size_change(self) -> None:
        self._check_number_of_bells()
        self._opening_row = generate_starting_row(
            self.number_of_bells, self.row_generator.custom_start_row)
        self._rounds = rounds(self.number_of_bells)
        self._check_starting_row()

    def _check_starting_row(self) -> bool:
        if (self.row_generator.custom_start_row is not None and
                len(self.row_generator.custom_start_row) < self.number_of_bells):
            self.logger.info(f"The starting row '{self.row_generator.custom_start_row}' "
                             + f"contains fewer bells than the tower ({self._tower.number_of_bells}). "
                             + "Wheatley will add the extra bells to the end of the change.")

        if len(self._opening_row) != self.number_of_bells:
            self.logger.warning(f"The current tower has fewer bells ({self._tower.number_of_bells}) "
                                + f"than the starting row {self._opening_row}. Wheatley will not ring!")
            return False
        return True

    def _check_number_of_bells(self) -> bool:
        """ Returns whether Wheatley can ring with the current number of bells with reasons why not """
        if self.row_generator.stage == 0:
            self.logger.debug("Place holder row generator. Wheatley will not ring!")
            return False
        if self._tower.number_of_bells < self.row_generator.stage:
            self.logger.warning(f"Row generation requires at least {self.row_generator.stage} bells, "
                                + f"but the current tower has {self._tower.number_of_bells}. "
                                + "Wheatley will not ring!")
            return False
        if self._tower.number_of_bells > self.row_generator.stage + 1 \
                and self.row_generator.custom_start_row is None:
            if self.row_generator.stage % 2:
                expected = self.row_generator.stage + 1
            else:
                expected = self.row_generator.stage
            self.logger.info(f"Current tower has more bells ({self._tower.number_of_bells}) than expected "
                             + f"({expected}). Wheatley will add extra cover bells.")
        return True

    def _on_look_to(self) -> None:
        if self._check_starting_row() and self._check_number_of_bells():
            self.look_to_has_been_called(time.time())
        # All Wheatley instances should return a 'Roll Call' message after `Look To` is called.
        if self._server_instance_id is not None:
            self._tower.emit_roll_call(self._server_instance_id)

    # This has to be made public, because the server's main function might have to call it
    def look_to_has_been_called(self, call_time: float) -> None:
        """ Callback called when a user calls 'Look To'. """
        self._rhythm.return_to_mainloop()

        treble = self._rounds[0]

        # Count number of user controlled bells
        number_of_user_controlled_bells = sum(1 for bell in self._rounds if self._user_assigned_bell(bell))

        self._rhythm.initialise_line(self.number_of_bells, self._user_assigned_bell(treble),
                                     call_time + 3, number_of_user_controlled_bells)

        # Move to the next row generator if it's defined
        self.row_generator = self.next_row_generator or self.row_generator
        self.next_row_generator = None

        # Clear all the flags and counters
        self._should_stand = False
        self._should_start_ringing_rounds = False
        # Set _rounds_left_before_method if we are ringing up-down-in (3 rounds for backstroke
        # start; 2 for handstroke)
        if not self._do_up_down_in:
            self._rounds_left_before_method = None
        elif self.row_generator.start_stroke().is_hand():
            self._rounds_left_before_method = 2
        else:
            self._rounds_left_before_method = 3

        # Reset the state, so that Wheatley starts by ringing rounds
        self._is_ringing = True
        self._is_ringing_rounds = True
        self._is_ringing_opening_row = True

        # Start at the first place of the first row
        self.start_next_row(is_first_row=True)

    def _on_go(self) -> None:
        """ Callback called when a user calls 'Go'. """
        if self._is_ringing_rounds:
            # Calculate how many more rows of rounds we should ring before going into changes (1 if
            # the person called 'Go' on the same stroke as the RowGenerator starts, otherwise 0).
            # These values are one less than expected because we are setting
            # _rounds_left_before_method **after** the row has started.
            self._rounds_left_before_method = 1 if self.stroke == self.row_generator.start_stroke() else 0
            # Make sure to call all of the calls that we have missed in the right order (in case the
            # person calling `Go` called it stupidly late)
            early_calls = [
                (ind, _calls)
                for (ind, _calls) in self.row_generator.early_calls().items()
                if ind > self._rounds_left_before_method
            ]
            # Sort early calls by the number of rows **before** the method start.  Note that we are
            # sorting by a quantity that counts **down** with time, hence the reversed sort.
            early_calls.sort(key=lambda x: x[0], reverse=True)
            # In this case, we don't want to wait until the next row before making these calls
            # because the rows on which these calls should have been called have already passed.
            # Therefore, we simply get them out as quickly as possible so they have the best chance
            # of being heard.
            for (_index, _calls) in early_calls:
                self._make_calls(_calls)

    def _on_bob(self) -> None:
        """ Callback called when a user calls 'Bob'. """
        self.row_generator.set_bob()

    def _on_single(self) -> None:
        """ Callback called when a user calls 'Single'. """
        self.row_generator.set_single()

    def _on_thats_all(self) -> None:
        """ Callback called when a user calls 'That`s All'. """
        self._should_start_ringing_rounds = True

    def _on_stand_next(self) -> None:
        """ Callback called when a user calls 'Stand Next'. """
        self._should_stand = True

    def _on_bell_ring(self, bell: Bell, stroke: Stroke) -> None:
        """ Callback called when the Tower receives a signal that a bell has been rung. """
        if self._user_assigned_bell(bell):
            # This will give us the stroke _after_ the bell rings, we have to invert it, because
            # otherwise this will always expect the bells on the wrong stroke and no ringing will
            # ever happen
            self._rhythm.on_bell_ring(bell, stroke.opposite(), time.time())

    def _on_stop_touch(self) -> None:
        self.logger.info("Got to callback for stop touch")
        self._tower.set_is_ringing(False)
        self._is_ringing = False
        self._rhythm.return_to_mainloop()

    # Mainloop and helper methods
    def expect_bell(self, index: int, bell: Bell) -> None:
        """ Called to let the rhythm expect a user-controlled bell at a certain time and stroke. """
        if self._user_assigned_bell(bell):
            self._rhythm.expect_bell(
                bell,
                self._row_number,
                index,
                self.stroke
            )

    def generate_next_row(self) -> None:
        """ Creates a new row from the row generator and tells the rhythm to expect the new bells. """
        if self._is_ringing_opening_row:
            self._row = self._opening_row
        elif self._is_ringing_rounds:
            self._row = self._rounds
        else:
            self._row, self._calls = self.row_generator.next_row_and_calls(self.stroke)
            # Add cover bells if needed
            if len(self._row) < len(self._opening_row):
                self._row = Row(self._row + self._opening_row[len(self._row):])

        bells = " ".join([str(bell) for bell in self._row])
        self.logger.info(f"ROW: {bells}")

    def start_next_row(self, is_first_row: bool) -> None:
        """ Updates state of bot ready to ring the next row / stop ringing"""
        # Generate the next row and update row indices
        self._place = 0
        if is_first_row:
            self._row_number = 0
        else:
            self._row_number += 1

        # Useful local variables
        has_just_rung_rounds = self._row == self._rounds
        next_stroke = Stroke.from_index(self._row_number)

        # Implement handbell-style stopping at rounds
        if self._stop_at_rounds and has_just_rung_rounds and not self._is_ringing_rounds:
            self._should_stand = False
            self._is_ringing = False

        # Set any early calls specified by the row generator to be called at the start of the next
        # row
        if self._rounds_left_before_method is not None:
            self._calls = self.row_generator.early_calls().get(self._rounds_left_before_method) or []

        # Start the method if necessary
        if self._rounds_left_before_method == 0:
            # Sanity check that we are in fact starting on the correct stroke (which is no longer
            # trivially guaranteed since we use a counter rather than a flag to determine when to
            # start the method)
            assert next_stroke == self.row_generator.start_stroke()
            self._rounds_left_before_method = None
            self._is_ringing_rounds = False
            self._is_ringing_opening_row = False
            # If the tower size somehow changed, then call 'Stand' but keep ringing rounds (Wheatley
            # calling 'Stand' will still generate a callback to `self._on_stand_next`, so we don't
            # need to handle that here)
            if not self._check_number_of_bells():
                self._make_call("Stand")
                self._is_ringing_rounds = True
            self.row_generator.reset()
        if self._rounds_left_before_method is not None:
            self._rounds_left_before_method -= 1

        # If we're starting a handstroke ...
        if next_stroke.is_hand():
            # ... and 'Stand' has been called, then stand
            if self._should_stand:
                self._should_stand = False
                self._is_ringing = False
            # ... and "That's All" has been called, then start ringing rounds.
            if self._should_start_ringing_rounds and not self._is_ringing_rounds:
                self._should_start_ringing_rounds = False
                self._is_ringing_rounds = True

        # If we've set `_is_ringing` to False, then no more rounds can happen so early return to
        # avoid erroneous calls
        if not self._is_ringing:
            return

        # Generate the next row, and tell the rhythm detection where the next row's bells are
        # expected to ring
        self.generate_next_row()
        for (index, bell) in enumerate(self._row):
            self.expect_bell(index, bell)

    def tick(self) -> None:
        """ Move the ringing on by one place """
        bell = self._row[self._place]
        user_controlled = self._user_assigned_bell(bell)

        self._rhythm.wait_for_bell_time(time.time(), bell, self._row_number, self._place,
                                        user_controlled, self.stroke)

        if not user_controlled:
            self._tower.ring_bell(bell, self.stroke)

        # If we are ringing the first bell in the row, then also make any calls that are needed.
        if self._place == 0:
            self._make_calls(self._calls)

        # Move one place through the ringing
        self._place += 1

        # Start a new row if we get to a place that's bigger than the number of bells
        if self._place >= self.number_of_bells:
            self.start_next_row(is_first_row=False)

    def main_loop(self) -> None:
        """
        Wheatley's main loop.  The main thread will get stuck forever in this function whilst
        Wheatley rings.
        """
        while True:
            # Log a message to say that Wheatley is waiting for 'Look To!'
            self.logger.info("Waiting for 'Look To'...")
            # Sit in an infinite loop whilst we're not ringing, and exit Wheatley if enough time
            # has passed
            self._last_activity_time = time.time()
            while not self._is_ringing:
                time.sleep(0.01)
                if self._server_mode and time.time() > self._last_activity_time + INACTIVITY_EXIT_TIME:
                    self.logger.info(f"Timed out - no activity for {INACTIVITY_EXIT_TIME}s. Exiting.")
                    return

            if self._do_up_down_in:
                self.logger.info(f"Starting to ring {self.row_generator.summary_string()}")
            else:
                self.logger.info(f"Waiting for 'Go' to ring {self.row_generator.summary_string()}...")

            if self._server_mode:
                self._tower.set_is_ringing(True)

            while self._is_ringing:
                self.tick()
                time.sleep(0.01)

            self.logger.info("Stopping ringing!")
            if self._server_mode:
                self._tower.set_is_ringing(False)

    def _user_assigned_bell(self, bell: Bell) -> bool:
        """ Returns `True` if this bell is not assigned to Wheatley. """
        return not self._bot_assigned_bell(bell)

    def _bot_assigned_bell(self, bell: Bell) -> bool:
        """ Returns `True` if this bell **is** assigned to Wheatley. """
        return self._tower.is_bell_assigned_to(bell, self._user_name)

    def _make_calls(self, call_list: List[str]) -> None:
        """ Broadcast a sequence of calls """
        for c in call_list:
            self._make_call(c)

    def _make_call(self, call: str) -> None:
        """ Broadcast a call, unless we've been told not to call anything. """
        if self._call_comps:
            self._tower.make_call(call)



import time
import unittest
from threading import Thread

from unittest.mock import Mock

from wheatley.stroke import HANDSTROKE, BACKSTROKE
from wheatley.bell import Bell
from wheatley.rhythm import Rhythm, WaitForUserRhythm

treble = Bell.from_number(1)
second = Bell.from_number(2)


class WaitForUserRhythmTests(unittest.TestCase):
    def setUp(self):
        self.mock_inner_rhythm = Mock(spec=Rhythm)
        self.wait_rhythm = WaitForUserRhythm(self.mock_inner_rhythm)
        self.wait_rhythm.sleep = self._patched_sleep

        self.waiting_for_bell_time = False

        self._finished_test = False
        self._sleeping = False
        self._return_from_sleep = False
        self.total_sleep_calls = 0
        self.total_sleep_delay = 0

    def tearDown(self):
        self._finished_test = True

    def _patched_sleep(self, seconds):
        """ Replacement sleep function that loops until advance_patched_sleep() is called
        Allows controlling how many times sleep() returns.
        """
        self._sleeping = True
        self.total_sleep_calls += 1
        self.total_sleep_delay += seconds
        while not self._return_from_sleep and not self._finished_test:
            time.sleep(0.001)
        self._return_from_sleep = False
        self._sleeping = False

    def advance_patched_sleep(self, seconds: float = 0.01):
        """ Makes _patched_sleep() return
        Allows controlling how many times sleep() returns.
        """
        expected_sleeps = seconds / WaitForUserRhythm.sleep_time
        for _ in range(round(expected_sleeps)):
            while not self._sleeping or self._return_from_sleep:
                time.sleep(0.001)
            self._return_from_sleep = True

    def start_wait_for_bell_time_thread(self, current_time, bell, row_number, place, user_controlled, stroke):
        """ Runs wait_for_bell_time() on a different thread, as it should loop until on_bell_ring() is called
        """

        def _wait_for_bell_time():
            self.waiting_for_bell_time = True
            self.wait_rhythm.wait_for_bell_time(current_time, bell, row_number, place, user_controlled,
                                                stroke)
            self.waiting_for_bell_time = False

        wait_thread = Thread(name=f"wait_for_bell_time_{bell}", target=_wait_for_bell_time)
        wait_thread.start()

    def assert_not_waiting_for_bell_time(self):
        count = 0
        # Return from a last sleep just in case
        self._return_from_sleep = True
        time.sleep(0.01)

        while self.waiting_for_bell_time and count < 50:
            time.sleep(0.001)
            count += 1
        self.assertFalse(self.waiting_for_bell_time, "Waiting for bell to ring")
        self._return_from_sleep = False

    def test_on_bell_ring__no_initial_delay(self):
        self.wait_rhythm.expect_bell(treble, 1, 1, HANDSTROKE)

        self.wait_rhythm.on_bell_ring(treble, HANDSTROKE, 0.5)

        self.mock_inner_rhythm.on_bell_ring.assert_called_once_with(treble, HANDSTROKE, 0.5)

    def test_on_bell_ring__subtracts_existing_delay(self):
        self.wait_rhythm.expect_bell(treble, 1, 1, HANDSTROKE)
        self.wait_rhythm.delay = 10

        self.wait_rhythm.on_bell_ring(treble, HANDSTROKE, 10.5)

        self.mock_inner_rhythm.on_bell_ring.assert_called_once_with(treble, HANDSTROKE, 0.5)

    def test_on_bell_ring__uses_original_delay_while_waiting(self):
        # Arrange
        initial_delay = 10
        expected_time = 11
        actual_time = 11.1

        self.wait_rhythm.expect_bell(treble, 1, 1, HANDSTROKE)
        self.wait_rhythm.delay = initial_delay

        # Start waiting for treble
        self.start_wait_for_bell_time_thread(current_time=expected_time, bell=treble, row_number=1, place=1,
                                             user_controlled=True, stroke=HANDSTROKE)

        self.advance_patched_sleep(seconds=actual_time - expected_time)

        # Treble rung by user
        self.wait_rhythm.on_bell_ring(treble, HANDSTROKE, actual_time)
        self.assert_not_waiting_for_bell_time()

        # 11.1 - 10 = 1.1
        self.mock_inner_rhythm.on_bell_ring.assert_called_once_with(treble, HANDSTROKE,
                                                                    actual_time - initial_delay)
        # 10 + (11.1 - 11) = 10.1
        self.assertEqual(10.1, round(self.wait_rhythm.delay, 2))

    def test_wait_for_bell_time__bell_rung_early_doesnt_wait(self):
        # Arrange
        initial_delay = 10
        expected_time = 11
        actual_time = 10.5

        self.wait_rhythm.delay = initial_delay
        self.wait_rhythm.expect_bell(treble, 1, 1, HANDSTROKE)

        # Treble rung by user
        self.wait_rhythm.on_bell_ring(treble, HANDSTROKE, actual_time)

        # Start waiting for treble
        self.start_wait_for_bell_time_thread(current_time=expected_time, bell=treble, row_number=1, place=1,
                                             user_controlled=True, stroke=HANDSTROKE)

        # Returns immediately without waiting
        self.assert_not_waiting_for_bell_time()

        self.mock_inner_rhythm.on_bell_ring.assert_called_once_with(treble, HANDSTROKE,
                                                                    actual_time - initial_delay)
        self.assertEqual(initial_delay, self.wait_rhythm.delay)

    def test_wait_for_bell_time__bell_rung_early_in_row_doesnt_wait(self):
        # Arrange
        initial_delay = 0
        expected_first_bell_time = 1
        expected_second_bell_time = 2

        self.wait_rhythm.expect_bell(treble, 1, 1, HANDSTROKE)
        self.wait_rhythm.expect_bell(second, 1, 2, HANDSTROKE)
        self.wait_rhythm.delay = initial_delay

        # Start waiting for treble
        self.start_wait_for_bell_time_thread(current_time=expected_first_bell_time, bell=treble, row_number=1,
                                             place=1, user_controlled=True, stroke=HANDSTROKE)

        self.advance_patched_sleep(seconds=0.05)

        # Second rung early by user
        self.wait_rhythm.on_bell_ring(second, HANDSTROKE, 1.05)

        self.advance_patched_sleep(seconds=0.05)
        self.assertTrue(self.waiting_for_bell_time)

        # Treble rung by user
        self.wait_rhythm.on_bell_ring(treble, HANDSTROKE, 1.1)
        self.assert_not_waiting_for_bell_time()

        # Start waiting for second
        self.start_wait_for_bell_time_thread(current_time=expected_second_bell_time, bell=second,
                                             row_number=1, place=1, user_controlled=True, stroke=HANDSTROKE)
        # Returns immediately without waiting
        self.assert_not_waiting_for_bell_time()

        self.mock_inner_rhythm.on_bell_ring.assert_any_call(second, HANDSTROKE, 1.05)
        self.mock_inner_rhythm.on_bell_ring.assert_any_call(treble, HANDSTROKE, 1.1)
        # 1.1 - 1, Second being early has not added delay
        self.assertEqual(0.1, round(self.wait_rhythm.delay, 2))

    def test_wait_for_bell_time__bell_rung_early_in_previous_row_doesnt_wait(self):
        # Arrange
        initial_delay = 0
        expected_first_time = 1
        expected_second_time = 2

        self.wait_rhythm.expect_bell(treble, 1, 1, HANDSTROKE)
        self.wait_rhythm.delay = initial_delay

        # Treble rung by user twice, so is now on wrong stroke
        self.wait_rhythm.on_bell_ring(treble, HANDSTROKE, 1)
        self.wait_rhythm.on_bell_ring(treble, BACKSTROKE, 1.1)

        self.start_wait_for_bell_time_thread(current_time=expected_first_time, bell=treble, row_number=1,
                                             place=1, user_controlled=True, stroke=HANDSTROKE)
        self.assert_not_waiting_for_bell_time()

        # Next row, treble is already rung BACKSTROKE
        self.wait_rhythm.expect_bell(treble, 2, 1, BACKSTROKE)

        self.start_wait_for_bell_time_thread(current_time=expected_second_time, bell=treble, row_number=2,
                                             place=1, user_controlled=True, stroke=BACKSTROKE)

        self.assert_not_waiting_for_bell_time()

        self.assertEqual(0, self.wait_rhythm.delay)

        self.mock_inner_rhythm.on_bell_ring.assert_any_call(treble, HANDSTROKE, 1)
        self.mock_inner_rhythm.on_bell_ring.assert_any_call(treble, BACKSTROKE, 1.1)

    def test_wait_for_bell_time__bell_early_and_corrected_in_previous_row_waits_for_bell(self):
        # Arrange
        expected_first_time = 1
        expected_second_time = 2

        self.wait_rhythm.expect_bell(treble, 1, 1, HANDSTROKE)

        # Treble rung by user
        self.wait_rhythm.on_bell_ring(treble, HANDSTROKE, 1)

        self.start_wait_for_bell_time_thread(current_time=expected_first_time, bell=treble, row_number=1,
                                             place=1, user_controlled=True, stroke=HANDSTROKE)

        # Treble accidentally rung by user and corrected back to the right stroke
        self.wait_rhythm.on_bell_ring(treble, BACKSTROKE, 1.1)
        self.wait_rhythm.on_bell_ring(treble, HANDSTROKE, 1.2)

        self.assert_not_waiting_for_bell_time()

        self.wait_rhythm.expect_bell(treble, 2, 1, BACKSTROKE)

        self.start_wait_for_bell_time_thread(current_time=expected_second_time, bell=treble, row_number=2,
                                             place=1, user_controlled=True, stroke=BACKSTROKE)

        self.assertTrue(self.waiting_for_bell_time)

        self.wait_rhythm.on_bell_ring(treble, BACKSTROKE, expected_second_time)

        self.assertEqual(0, self.wait_rhythm.delay)

        self.mock_inner_rhythm.on_bell_ring.assert_any_call(treble, HANDSTROKE, 1)
        self.mock_inner_rhythm.on_bell_ring.assert_any_call(treble, BACKSTROKE, 1.1)
        self.mock_inner_rhythm.on_bell_ring.assert_any_call(treble, HANDSTROKE, 1.2)
        self.mock_inner_rhythm.on_bell_ring.assert_any_call(treble, BACKSTROKE, expected_second_time)


if __name__ == '__main__':
    unittest.main()



import unittest
from wheatley.bell import Bell, BELL_NAMES, MAX_BELL


class BellTests(unittest.TestCase):
    def test_to_from_string_success(self):
        for name in BELL_NAMES:
            with self.subTest(name=name):
                self.assertEqual(str(Bell.from_str(name)), name)

    def test_from_string_error(self):
        for name in ['p', 'a', 'x', 'O', '?', '14', 'q1', '00']:
            with self.subTest(name=name):
                with self.assertRaises(ValueError):
                    Bell.from_str(name)

    def test_from_index_success(self):
        for i in range(MAX_BELL):
            self.assertEqual(Bell.from_index(i).index, i)

    def test_from_index_error(self):
        for i in [-1, MAX_BELL, MAX_BELL + 1, 100, -1100]:
            with self.assertRaises(ValueError):
                Bell.from_index(i)
        # This should be disallowed by the type checker, but may as well test that it does fail
        with self.assertRaises(TypeError):
            Bell.from_index(None)

    def test_from_number_success(self):
        for i in range(1, MAX_BELL + 1):
            self.assertEqual(Bell.from_number(i).number, i)

    def test_from_number_error(self):
        for i in [-1, 0, MAX_BELL + 1, MAX_BELL + 2, 100, -1100]:
            with self.assertRaises(ValueError):
                Bell.from_number(i)
        # This should be disallowed by the type checker, but may as well test that it does fail
        with self.assertRaises(TypeError):
            Bell.from_number(None)

    def test_equality(self):
        for i in range(MAX_BELL):
            for j in range(MAX_BELL):
                # Bells with indices `i` and `j` should be equal precisely when `i == j`
                self.assertEqual(Bell.from_index(i) == Bell.from_index(j), i == j)
                # Bells with indices `i` and `j` should be not equal precisely when `i != j`
                self.assertEqual(Bell.from_index(i) != Bell.from_index(j), i != j)


if __name__ == '__main__':
    unittest.main()



"""
Module that provides a convenient interface between Ringing Room and the rest of the program.
"""

import collections
import logging
from time import sleep
from typing import Optional, Callable, Dict, List, Any

import socketio # type: ignore

from wheatley.aliases import JSON
from wheatley.bell import Bell
from wheatley.stroke import Stroke, HANDSTROKE


class RingingRoomTower:
    """ A class representing a tower, which will handle a single ringing-room session. """

    logger_name = "TOWER"

    def __init__(self, tower_id: int, url: str) -> None:
        """ Initialise a tower with a given room id and url. """
        self.tower_id = tower_id

        self._bell_state: List[Stroke] = []
        self._assigned_users: Dict[Bell, int] = {}
        # A map from user IDs to the corresponding user name
        self._user_name_map: Dict[int, str] = {}

        self.invoke_on_call: Dict[str, List[Callable[[], Any]]] = collections.defaultdict(list)
        self.invoke_on_reset: List[Callable[[], Any]] = []
        self.invoke_on_bell_rung: List[Callable[[Bell, Stroke], Any]] = []
        self.invoke_on_setting_change: List[Callable[[str, Any], Any]] = []
        self.invoke_on_row_gen_change: List[Callable[[Any], Any]] = []
        self.invoke_on_stop_touch: List[Callable[[], Any]] = []

        self._url = url
        self._socket_io_client: Optional[socketio.Client] = None

        self.logger = logging.getLogger(self.logger_name)

    def __enter__(self) -> Any:
        """ Called when entering a 'with' block.  Opens the socket-io connection. """
        self.logger.debug("ENTER")

        if self._socket_io_client is not None:
            raise Exception("Trying to connect twice")

        self._create_client()

        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        """
        Called when finishing a 'with' block.  Clears up the object and disconnects the session.
        """
        self.logger.debug("EXIT")
        if self._socket_io_client:
            self.logger.info("Disconnect")
            self._socket_io_client.disconnect()
            self._socket_io_client = None

    @property
    def number_of_bells(self) -> int:
        """ Returns the number of bells currently in the tower. """
        return len(self._bell_state)

    def ring_bell(self, bell: Bell, expected_stroke: Stroke) -> bool:
        """ Send a request to the the server if the bell can be rung on the given stroke. """
        try:
            stroke = self.get_stroke(bell)
            if stroke != expected_stroke:
                self.logger.error(f"Bell {bell} on opposite stroke")
                return False

            bell_num: int = bell.number
            is_handstroke: bool = stroke.is_hand()
            self._emit("c_bell_rung", {"bell": bell_num, "stroke": is_handstroke, "tower_id": self.tower_id})

            return True
        except Exception as e:
            self.logger.error(e)
            return False

    def is_bell_assigned_to(self, bell: Bell, user_name: Optional[str]) -> bool:
        """ Returns true if a given bell is assigned to the given user name. """
        try:
            assigned_user_id = self._assigned_users.get(bell, None)
            if assigned_user_id is None:
                return user_name is None
            return self.user_name_from_id(assigned_user_id) == user_name
        except KeyError:
            return False

    def user_name_from_id(self, user_id: int) -> Optional[str]:
        """
        Converts a numerical user ID into the corresponding user name, returning None if user_id is not in
        the tower.
        """
        return self._user_name_map.get(user_id)

    def get_stroke(self, bell: Bell) -> Optional[Stroke]:
        """ Returns the stroke of a given bell. """
        if bell.index >= len(self._bell_state) or bell.index < 0:
            self.logger.error(f"Bell {bell} not in tower")
            return None
        return self._bell_state[bell.index]

    def make_call(self, call: str) -> None:
        """ Broadcasts a given call to the other users of the tower. """
        self.logger.info(f"EMIT: Calling '{call}'")
        self._emit("c_call", {"call": call, "tower_id": self.tower_id})

    def set_at_hand(self) -> None:
        """ Sets all the bells at hand. """
        self.logger.info("EMIT: Setting bells at handstroke")
        self._emit("c_set_bells", {"tower_id": self.tower_id})

    def set_number_of_bells(self, number: int) -> None:
        """ Set the number of bells in the tower. """
        self.logger.info(f"EMIT: Setting size to {number}")
        self._emit("c_size_change", {"new_size": number, "tower_id": self.tower_id})

    def set_is_ringing(self, value: bool) -> None:
        """ Broadcast to all the users that Wheatley has started or stopped ringing. """
        self.logger.info(f"EMIT: Telling RR clients to set is_ringing to {value}")
        self._emit("c_wheatley_is_ringing", {"is_ringing": value, "tower_id": self.tower_id})

    def emit_roll_call(self, instance_id: int) -> None:
        """ Broadcasts a 'roll call' reply to the Ringing Room server. """
        self.logger.info("EMIT: Replying to roll call")
        self._emit("c_roll_call", {"tower_id": self.tower_id, "instance_id": instance_id})

    def wait_loaded(self) -> None:
        """ Pause the thread until the socket-io connection is open and stable. """
        if self._socket_io_client is None or not self._socket_io_client.connected:
            raise SocketIOClientError("Not Connected")

        iteration = 0
        # Wait up to 2 seconds
        while iteration < 20:
            if self._bell_state:
                break

            iteration += 1
            sleep(0.1)
        else:
            raise SocketIOClientError("Not received bell state from RingingRoom")

    def _create_client(self) -> None:
        """ Generates the socket-io client and attaches callbacks. """
        self._socket_io_client = socketio.Client()
        self._socket_io_client.connect(self._url)
        self.logger.debug(f"Connected to {self._url}")

        self._socket_io_client.on("s_bell_rung", self._on_bell_rung)
        self._socket_io_client.on("s_global_state", self._on_global_bell_state)
        self._socket_io_client.on("s_user_entered", self._on_user_entered)
        self._socket_io_client.on("s_set_userlist", self._on_user_list)
        self._socket_io_client.on("s_size_change", self._on_size_change)
        self._socket_io_client.on("s_assign_user", self._on_assign_user)
        self._socket_io_client.on("s_call", self._on_call)
        self._socket_io_client.on("s_user_left", self._on_user_leave)
        self._socket_io_client.on("s_wheatley_setting", self._on_setting_change)
        self._socket_io_client.on("s_wheatley_row_gen", self._on_row_gen_change)
        self._socket_io_client.on("s_wheatley_stop_touch", self._on_stop_touch)

        self._join_tower()
        self._request_global_state()

    # The values in data could have any types, so we don't need any type checking here.
    def _on_setting_change(self, data: JSON) -> None:
        # Log a message (to info if the setting change is used, debug otherwise)
        is_ignored = len(self.invoke_on_setting_change) == 0
        log_message = f"RECEIVED: Settings changed: {data}" \
            + "{' (ignoring)' if is_ignored else ''}"
        if is_ignored:
            self.logger.debug(log_message)
        else:
            self.logger.info(log_message)

        for key, value in data.items():
            for callback in self.invoke_on_setting_change:
                callback(key, value)

    def _on_row_gen_change(self, data: JSON) -> None:
        # Log a message (to info if the setting change is used, debug otherwise)
        is_ignored = len(self.invoke_on_row_gen_change) == 0
        log_message = f"RECEIVED: Row gen changed: {data}" \
            + "{' (ignoring)' if is_ignored else ''}"
        if is_ignored:
            self.logger.debug(log_message)
        else:
            self.logger.info(log_message)

        for callback in self.invoke_on_row_gen_change:
            callback(data)

    def _on_stop_touch(self, data: JSON) -> None:
        # No data is transferred over this signal, so we don't need to typecheck `data`
        self.logger.info(f"RECEIVED: Stop touch: {data}.")

        for callback in self.invoke_on_stop_touch:
            callback()

    def _on_user_leave(self, data: JSON) -> None:
        # Unpack the data and assign it the expected types
        user_id_that_left: int = data["user_id"]
        user_name_that_left: str = data["username"]

        # Remove the user ID that left from our user list
        if user_id_that_left not in self._user_name_map:
            self.logger.warning(
                f"User #{user_id_that_left}:'{user_name_that_left}' left, but wasn't in the user list."
            )
        elif self._user_name_map[user_id_that_left] != user_name_that_left:
            self.logger.warning(f"User #{user_id_that_left}:'{user_name_that_left}' left, but that ID was \
logged in as '{self._user_name_map[user_id_that_left]}'.")
            del self._user_name_map[user_id_that_left]

        bells_unassigned: List[Bell] = []

        # Unassign all instances of that user
        for bell, user in self._assigned_users.items():
            if user == user_id_that_left:
                bells_unassigned.append(bell)
        for bell in bells_unassigned:
            del self._assigned_users[bell]

        self.logger.info(
            f"RECEIVED: User #{user_id_that_left}:'{user_name_that_left}' left from bells {bells_unassigned}."
        )

    def _on_user_entered(self, data: JSON) -> None:
        """ Called when the server receives a uew user so we can update our user list. """
        # Unpack the data and assign it expected types
        user_id: int = data['user_id']
        username: str = data['username']
        # Add the new user to the user list, so we can match up their ID with their username
        self._user_name_map[user_id] = username

    def _on_user_list(self, user_list: JSON) -> None:
        """ Called when the server broadcasts a user list when Wheatley joins a tower. """
        for user in user_list['user_list']:
            # Unpack the data and assign it expected types
            user_id: int = user['user_id']
            username: str = user['username']
            self._user_name_map[user_id] = username

    def _join_tower(self) -> None:
        """ Joins the tower as an anonymous user. """
        self.logger.info(f"EMIT: Joining tower {self.tower_id}")
        self._emit(
            "c_join",
            {"anonymous_user": True, "tower_id": self.tower_id},
        )

    def _request_global_state(self) -> None:
        """ Send a request to the server to get the current state of the tower. """
        self.logger.debug("EMIT: Requesting global state.")
        self._emit('c_request_global_state', {"tower_id": self.tower_id})

    def _on_bell_rung(self, data: JSON) -> None:
        """ Callback called when the client receives a signal that a bell has been rung. """
        # Unpack the data and assign it expected types
        global_bell_state: List[bool] = data['global_bell_state']
        who_rang_raw: int = data["who_rang"]
        # Run callbacks for updating the bell state
        self._update_bell_state([Stroke(b) for b in global_bell_state])
        # Convert 'who_rang' to a Bell
        who_rang = Bell.from_number(who_rang_raw)
        # Only run the callbacks if the bells exist
        for bell_ring_callback in self.invoke_on_bell_rung:
            new_stroke = self.get_stroke(who_rang)
            if new_stroke is None:
                self.logger.warning(
                    f"Bell {who_rang} rang, but the tower only has {self.number_of_bells} bells."
                )
            else:
                bell_ring_callback(who_rang, new_stroke)

    def _update_bell_state(self, bell_state: List[Stroke]) -> None:
        self._bell_state = bell_state
        self.logger.debug(f"RECEIVED: Bells '{''.join([s.char() for s in bell_state])}'")

    def _on_global_bell_state(self, data: JSON) -> None:
        """
        Callback called when receiving an update to the global tower state.
        Cannot have further callbacks assigned to it.
        """
        global_bell_state: List[bool] = data["global_bell_state"]
        self._update_bell_state([Stroke(x) for x in global_bell_state])
        for invoke_callback in self.invoke_on_reset:
            invoke_callback()

    def _on_size_change(self, data: JSON) -> None:
        """ Callback called when the number of bells in the room changes. """
        new_size: int = data["size"]
        if new_size != self.number_of_bells:
            # Remove the user who's bells have been removed (so that returning to a stage doesn't make
            # Wheatley think the bells are still assigned)
            self._assigned_users = {bell: user for (bell, user) in self._assigned_users.items()
                                    if bell.number <= new_size}
            # Set the bells at handstroke
            self._bell_state = self._bells_set_at_hand(new_size)
            # Handle all the callbacks
            self.logger.info(f"RECEIVED: New tower size '{new_size}'")
            for invoke_callback in self.invoke_on_reset:
                invoke_callback()

    def _on_assign_user(self, data: JSON) -> None:
        """ Callback called when a bell assignment is changed. """
        raw_bell: int = data["bell"]
        bell: Bell = Bell.from_number(raw_bell)
        user: Optional[int] = data["user"] or None

        assert isinstance(user, int) or user is None, \
               f"User ID {user} is not an integer (it has type {type(user)})."

        if user is None:
            self.logger.info(f"RECEIVED: Unassigned bell '{bell}'")
            if bell in self._assigned_users:
                del self._assigned_users[bell]
        else:
            self._assigned_users[bell] = user
            self.logger.info(f"RECEIVED: Assigned bell '{bell}' to '{self.user_name_from_id(user)}'")

    def _on_call(self, data: Dict[str, str]) -> None:
        """ Callback called when a call is made. """
        call = data["call"]
        self.logger.info(f"RECEIVED: Call '{call}'")

        found_callback = False
        for call_callback in self.invoke_on_call.get(call, []):
            call_callback()
            found_callback = True
        if not found_callback:
            self.logger.debug(f"No callback found for '{call}'")

    def _emit(self, event: str, data: Any) -> None:
        """ Emit a socket-io signal. """
        if self._socket_io_client is None or not self._socket_io_client.connected:
            raise SocketIOClientError("Not Connected")

        self._socket_io_client.emit(event, data)

    @staticmethod
    def _bells_set_at_hand(number: int) -> List[Stroke]:
        """ Returns the representation of `number` bells, all set at handstroke. """
        return [HANDSTROKE for _ in range(number)]


class SocketIOClientError(Exception):
    """Errors related to SocketIO Client"""



""" A module to hold common type aliases """

from typing import Any, Dict, NewType, List
from wheatley.bell import Bell


CallDef = NewType('CallDef', Dict[int, str])
JSON = Dict[str, Any]
Places = NewType('Places', List[int])
Row = NewType('Row', List[Bell])



""" Run a fuzzer to generate lots of random input for the peal speed parser. """

from random import choice, randint

from wheatley.parsing import parse_peal_speed, PealSpeedParseError

from .fuzz_utils import fuzz_for_unwrapped_errors


def random_peal_speed_string():
    """ Generate a plausible input value for `wheatley.arg_parsing.parse_peal_speed` to parse. """

    return "".join([choice("1234567890hm,\nx?ET ") for _ in range(randint(0, 20))])


def fuzz_parse_peal_speed():
    fuzz_for_unwrapped_errors(
        "parse_peal_speed",
        parse_peal_speed,
        random_peal_speed_string,
        PealSpeedParseError
    )



#!/usr/bin/env python3

"""
The file containing the main function for Wheatley, as well as all the command line argument parsing
required to make Wheatley easily configurable.
"""

import argparse
import logging
import os
import sys

from typing import List, Optional

from wheatley.rhythm import Rhythm, RegressionRhythm, WaitForUserRhythm
from wheatley.tower import RingingRoomTower
from wheatley.bot import Bot
from wheatley.page_parser import get_load_balancing_url, TowerNotFoundError, InvalidURLError
from wheatley.parsing import parse_peal_speed, PealSpeedParseError, parse_call, parse_start_row, \
                             StartRowParseError, parse_place_notation, PlaceNotationError

from wheatley.row_generation import RowGenerator, ComplibCompositionGenerator
from wheatley.row_generation import MethodPlaceNotationGenerator
from wheatley.row_generation.complib_composition_generator import PrivateCompError, InvalidCompError
from wheatley.row_generation.method_place_notation_generator import MethodNotFoundError, \
                                                                    generator_from_special_title
from wheatley.row_generation.place_holder_generator import PlaceHolderGenerator
from wheatley.row_generation.place_notation_generator import PlaceNotationGenerator


def create_row_generator(args: argparse.Namespace) -> RowGenerator:
    """ Generates a row generator according to the given CLI arguments. """
    if "comp" in args and args.comp is not None and "start_row" in args and args.start_row is not None:
        sys.exit("You may not specify a custom start row with a composition")
    elif "comp" in args and args.comp is not None:
        try:
            return ComplibCompositionGenerator.from_arg(args.comp)
        except (PrivateCompError, InvalidCompError) as e:
            sys.exit(f"Bad value for '--comp': {e}")
    elif "method" in args and args.method is not None:
        try:
            return generator_from_special_title(args.method, args.start_row) or \
                MethodPlaceNotationGenerator(
                    args.method,
                    parse_call(args.bob),
                    parse_call(args.single),
                    args.start_row,
                    args.start_index
                )
        except MethodNotFoundError as e:
            sys.exit(f"Bad value for '--method': {e}")
    elif "place_notation" in args and args.place_notation is not None:
        try:
            stage, place_notation = parse_place_notation(args.place_notation)
            return PlaceNotationGenerator(
                stage,
                place_notation,
                parse_call(args.bob),
                parse_call(args.single),
                args.start_index
            )
        except PlaceNotationError as e:
            sys.exit(f"Bad value for '--place-notation': {e}")

    raise AssertionError("One of --method, --comp or --place-notation should always be defined")


def create_rhythm(peal_speed: int, inertia: float, max_bells_in_dataset: int, handstroke_gap: float,
                  use_wait: bool, initial_inertia: float = 0) -> Rhythm:
    """ Generates a rhythm object according to the given CLI arguments. """
    # Sets the minimum number of bells that Wheatley will use in order to deduce a rhythm.  Setting this to
    # larger numbers will make Wheatley more stable during the pull-off, whereas smaller numbers will make
    # Wheatley more sensitive to user's pull-off speed.  If this is larger than `--max_bells_in_dataset`,
    # this will be clamped to fit.
    min_bells_in_dataset = 4

    # Clamp min_bells_in_dataset to not be bigger than max_bells_in_dataset
    min_bells_in_dataset = min(min_bells_in_dataset, max_bells_in_dataset)

    regression = RegressionRhythm(
        inertia,
        handstroke_gap=handstroke_gap,
        peal_speed=peal_speed,
        min_bells_in_dataset=min_bells_in_dataset,
        max_bells_in_dataset=max_bells_in_dataset,
        initial_inertia=initial_inertia
    )

    if use_wait:
        return WaitForUserRhythm(regression)

    return regression


def get_version_number() -> str:
    """
    Try to read the file with the version number, if not print an error and set a poison value as the
    version.
    """
    try:
        version_file_path = os.path.join(os.path.split(__file__)[0], "version.txt")

        with open(version_file_path) as f:
            return f.read()
    except IOError:
        return "<NO VERSION FILE FOUND>"


def configure_logging(verbose: bool, quietness: int) -> None:
    """ Sets up the logging for Wheatley. """
    logging.basicConfig(level=logging.WARNING)

    # Verboseness is a numerical represntation of the log level, with the following cases:
    # >= 1 -> DEBUG
    #    0 -> INFO
    #   -1 -> WARNING
    #   -2 -> ERROR
    # < -2 -> CRITICAL (Wheatley doesn't emit any CRITICAL logs, so this is essentially silent)
    verboseness: int = (1 if verbose else 0) - quietness

    if verboseness >= 1:
        log_level = logging.DEBUG
    elif verboseness == 0:
        log_level = logging.INFO
    elif verboseness == -1:
        log_level = logging.WARNING
    elif verboseness == -2:
        log_level = logging.ERROR
    else:  # verboseness <= -3
        log_level = logging.CRITICAL

    logging.getLogger(Bot.logger_name).setLevel(log_level)
    logging.getLogger(RingingRoomTower.logger_name).setLevel(log_level)
    logging.getLogger(RowGenerator.logger_name).setLevel(log_level)
    logging.getLogger(RegressionRhythm.logger_name).setLevel(log_level)
    logging.getLogger(WaitForUserRhythm.logger_name).setLevel(log_level)


def server_main(override_args: Optional[List[str]], stop_on_join_tower: bool) -> None:
    """
    The main function of Wheatley when spawned by the Ringing Room server.
    This has many many fewer options than the standard `main` function, on the basis that this running mode
    is designed to take its parameters over SocketIO whilst running, rather than from the CLI args.
    """
    __version__ = get_version_number()

    parser = argparse.ArgumentParser(
        description="A bot to fill in bells during ringingroom.com practices (server mode)."
    )

    parser.add_argument(
        "room_id",
        type=int,
        help="The numerical ID of the tower to join, represented as a row on 9 bells, \
              e.g. 763451928."
    )
    parser.add_argument(
        "-p", "--port",
        type=int,
        help="The port of the SocketIO server (which must be hosted on localhost)."
    )
    parser.add_argument(
        "-l", "--look-to-time",
        type=float,
        help="Set to the time when 'Look to' was called if Wheatley was spawned because 'look to' was \
              called and Wheatley is needed."
    )
    parser.add_argument(
        "-i", "--id",
        type=int,
        help="The instance ID of this Wheatley process."
    )
    parser.add_argument(
        "-v", "--verbose",
        action="count",
        default=0,
        help="Makes Wheatley print more (DEBUG) output."
    )
    parser.add_argument(
        "-q", "--quiet",
        action="count",
        default=0,
        help="Makes Wheatley print less output.  `-q` only prints WARNINGs and ERRORs; `-qq` only prints \
              ERRORs; `-qqq` prints nothing."
    )

    # Parse arguments
    # `[1:]` is apparently needed, because sys.argv[0] is the working file of the Python interpreter
    # which `parser.parse_args` does not want to see as an argument
    args = parser.parse_args(sys.argv[1:] if override_args is None else override_args)

    # Run the program
    configure_logging(args.verbose, args.quiet)

    # Log the version string to 'DEBUG'
    logging.debug(f"Running Wheatley {__version__}")

    # Args that we are currently 'missing'
    use_up_down_in = True
    stop_at_rounds = True
    peal_speed = 180
    inertia = 1
    initial_inertia = 0
    max_bells_in_dataset = 15
    handstroke_gap = 1
    use_wait = True
    call_comps = True

    tower_url = "http://127.0.0.1:" + str(args.port)

    tower = RingingRoomTower(args.room_id, tower_url)
    rhythm = create_rhythm(peal_speed, inertia, max_bells_in_dataset, handstroke_gap, use_wait,
                           initial_inertia)
    bot = Bot(tower, PlaceHolderGenerator(), use_up_down_in, stop_at_rounds, call_comps, rhythm,
              user_name="Wheatley", server_instance_id=args.id)

    with tower:
        tower.wait_loaded()

        if args.look_to_time is not None:
            bot.look_to_has_been_called(args.look_to_time)

        if not stop_on_join_tower:
            bot.main_loop()


def console_main(override_args: Optional[List[str]], stop_on_join_tower: bool) -> None:
    """
    The main function of Wheatley, when called from the Command Line.
    This parses the CLI arguments, creates the Rhythm, RowGenerator and Bot objects, then starts
    Wheatley's mainloop.

    The two optional arguments are used by the integration tester to give Wheatley artificial argument values
    (override_args) and make Wheatley exit with error code 0 on joining a tower so that hanging forever
    can be differentiated from Wheatley's normal behaviour of sitting in an infinite loop waiting for input.
    """
    __version__ = get_version_number()

    # PARSE THE ARGUMENTS

    parser = argparse.ArgumentParser(
        description="A bot to fill in bells during ringingroom.com practices"
    )

    # Tower arguments
    tower_group = parser.add_argument_group("Tower arguments")
    tower_group.add_argument(
        "room_id",
        type=int,
        help="The numerical ID of the tower to join, represented as a row on 9 bells, \
              e.g. 763451928."
    )
    tower_group.add_argument(
        "--url",
        default="https://ringingroom.com",
        type=str,
        help="The URL of the server to join (defaults to 'https://ringingroom.com')"
    )
    tower_group.add_argument(
        "--name",
        default=None,
        type=str,
        help="If set, then Wheatley will ring bells assigned to the given name. \
             When not set, Wheatley rings unassigned bells."
    )

    # Row generation arguments
    row_gen_group = parser.add_argument_group("Row generation arguments")

    # An mutual exclusion group to disallow specifying more than one of a method,
    # a CompLib comp or a place notation
    comp_method_place_group = row_gen_group.add_mutually_exclusive_group(required=True)
    comp_method_place_group.add_argument(
        "-c", "--comp",
        type=str,
        help="The ID or URL of the complib composition you want to ring"
    )
    comp_method_place_group.add_argument(
        "-m", "--method",
        type=str,
        help="The title of the method you want to ring"
    )
    comp_method_place_group.add_argument(
        "-p", "--place-notation",
        type=str,
        help="The place notation description of the method you want to ring"
    )

    row_gen_group.add_argument(
        "-b",
        "--bob",
        default="14",
        help='An override for what place notation(s) should be made when a `Bob` is called in \
              Ringing Room.  These will by default happen at the lead end.  Examples: "16" or \
              "0:16" => 6ths place lead end bob.  "-1:3" or "-1:3.1" => a Grandsire Bob.  "20: 70" \
              => a 70 bob taking effect 20 changes into a lead (the Half Lead for Surprise Royal). \
              "20:7/0:4" => a 70 bob 20 changes into a lead and a 14 bob at the lead end. \
              "3: 5/9: 5" => bobs in Stedman Triples.  Defaults to "14".'
    )
    row_gen_group.add_argument(
        "-n",
        "--single",
        default="1234",
        help='An override for what place notation(s) should be made when a `Single` is called in \
              Ringing Room.  These will by default happen at the lead end.  Examples: "1678" or \
              "0:1678" => 6ths place lead end single.  "-1:3.123" => a Grandsire Single. \
              "20: 7890" => a 7890 single taking effect 20 changes into a lead (the Half Lead for \
              Surprise Royal). "3: 567/9: 567" => singles in Stedman Triples.  Defaults to "1234".'
    )
    row_gen_group.add_argument(
        "--start-index",
        type=int,
        default=0,
        help="Determines which row of the lead Wheatley will start ringing.  This can be negative (so -1 \
              would refer to the lead **end**).  Defaults to 0 (i.e. a standard start)."
    )
    row_gen_group.add_argument(
        "--start-row",
        type=str,
        help="Determines the initial row."
    )
    row_gen_group.add_argument(
        "-u", "--use-up-down-in",
        action="store_true",
        help="If set, then the Wheatley will automatically go into changes after two rounds have been \
              rung."
    )
    row_gen_group.add_argument(
        "-s", "--stop-at-rounds",
        action="store_true",
        help="If set, then Wheatley will stand his bells the first time rounds is reached."
    )
    row_gen_group.add_argument(
        "-H", "--handbell-style",
        action="store_true",
        help="If set, then Wheatley will ring 'handbell style', i.e. ringing two strokes of \
              rounds then straight into changes, and stopping at the first set of rounds. By \
              default, he will ring 'towerbell style', i.e. only taking instructions from the \
              ringing-room calls. This is equivalent to using the '-us' flags."
    )
    row_gen_group.add_argument(
        "--no-calls",
        action="store_true",
        help="If set, Wheatley will not call anything when ringing compositions."
    )

    # Rhythm arguments
    rhythm_group = parser.add_argument_group("Rhythm arguments")
    rhythm_group.add_argument(
        "-k", "--keep-going",
        action="store_true",
        help="If set, Wheatley will not wait for users to ring - instead, he will push on with the \
              rhythm."
    )
    rhythm_group.add_argument(
        "-w", "--wait",
        action="store_true",
        help="Legacy parameter, which is now set by default. The previous default behaviour of not waiting \
              can be set with '-k'/'--keep-going'."
    )
    rhythm_group.add_argument(
        "-I", "--inertia",
        type=float,
        default=0.5,
        help="Overrides Wheatley's 'inertia' - how much Wheatley will take other ringers' positions \
              into account when deciding when to ring.  0.0 means he will cling as closely as \
              possible to the current rhythm, 1.0 means that he will completely ignore the other \
              ringers."
    )
    rhythm_group.add_argument(
        "-S", "--peal-speed",
        default="2h58",
        help="Sets the default speed that Wheatley will ring (assuming a peal of 5040 changes), \
              though this will usually be adjusted by Wheatley whilst ringing to keep with other \
              ringers.  Example formatting: '3h4' = '3h4m' = '3h04m' = '3h04' = '184m' = '184'. \
              Defaults to '2h58'."
    )
    rhythm_group.add_argument(
        "-G", "--handstroke-gap",
        type=float,
        default=1.0,
        help="Sets the handstroke gap as a factor of the space between two bells.  Defaults to \
              '1.0'."
    )
    rhythm_group.add_argument(
        "-X", "--max-bells-in-dataset",
        type=int,
        default=15,
        help="Sets the maximum number of bells that Wheatley will store to determine the current \
              ringing speed.  If you make this larger, then he will be more consistent but less \
              quick to respond to changes in rhythm.  Setting both this and \
              --inertia to a very small values could result in Wheatley ringing ridiculously \
              quickly.  Defaults to '15'."
    )

    # Misc arguments
    parser.add_argument(
        "--version",
        action="version",
        version=f"Wheatley {__version__}"
    )
    parser.add_argument(
        "-v", "--verbose",
        action="count",
        default=0,
        help="Makes Wheatley print more (DEBUG) output."
    )
    parser.add_argument(
        "-q", "--quiet",
        action="count",
        default=0,
        help="Makes Wheatley print less output.  `-q` only prints WARNINGs and ERRORs; `-qq` only prints \
              ERRORs; `-qqq` prints nothing."
    )


    # Parse arguments
    # `[1:]` is apparently needed, because sys.argv[0] is the working file of the Python interpreter
    # which `parser.parse_args` does not want to see as an argument
    args = parser.parse_args(sys.argv[1:] if override_args is None else override_args)

    # Deprecation warnings
    if args.wait:
        print("Deprecation warning: `--wait` has been replaced with `--keep-going`!")

    # Run the program
    configure_logging(args.verbose, args.quiet)

    try:
        tower_url = get_load_balancing_url(args.room_id, args.url)
    except TowerNotFoundError as e:
        sys.exit(f"Bad value for 'room_id': {e}")
    except InvalidURLError as e:
        sys.exit(f"Bad value for '--url': {e}")

    try:
        parse_start_row(args.start_row)
    except StartRowParseError as e:
        sys.exit(f"{e}")

    tower = RingingRoomTower(args.room_id, tower_url)
    row_generator = create_row_generator(args)

    try:
        peal_speed = parse_peal_speed(args.peal_speed)
    except PealSpeedParseError as e:
        sys.exit(f"{e}")

    rhythm = create_rhythm(peal_speed, args.inertia, args.max_bells_in_dataset, args.handstroke_gap,
                           not args.keep_going)
    bot = Bot(tower, row_generator, args.use_up_down_in or args.handbell_style,
              args.stop_at_rounds or args.handbell_style, not args.no_calls, rhythm, user_name=args.name)

    # Catch keyboard interrupts and just print 'Bye!' instead a load of guff
    try:
        with tower:
            tower.wait_loaded()
            if not stop_on_join_tower:
                bot.main_loop()
    except KeyboardInterrupt:
        print("Bye!")

def main(override_args: Optional[List[str]] = None, stop_on_join_tower: bool = False) -> None:
    """
    The root main function for Wheatley.  If the first argument given to Wheatley is 'server-mode', then
    this will run `server_main` (the main function for running Wheatley on RR's servers) otherwise it will
    run `console_main` (the main function for running Wheatley from the Command Line).
    """
    # If the user adds 'server-mode' to the command, run the server-mode main function instead of this one,
    # and remove the 'server-mode' argument
    unparsed_args = override_args or sys.argv[1:]
    if len(unparsed_args) > 1 and unparsed_args[0] == "server-mode":
        del unparsed_args[0]
        server_main(unparsed_args, stop_on_join_tower)
        # The `server_main` function shouldn't return, but if it does somehow manage to do so return instead
        # of running the standard main function as well
    else:
        console_main(unparsed_args, stop_on_join_tower)



if __name__ == "__main__":
    main()



"""
A module to hold a place-holder row generator that accepts all tower sizes but throws an exception instead of
producing rows.
"""

from wheatley.aliases import Row
from wheatley.stroke import Stroke

from .row_generator import RowGenerator


class NullRowGenError(Exception):
    """ An exception thrown when a row is requested from a 'NullGenerator'. """

    def __str__(self) -> str:
        return "_gen_row() called on a PlaceHolderGenerator"


class PlaceHolderGenerator(RowGenerator):
    """
    A place holder row generator that accepts any tower size but throws an exception when rows are requested.
    """

    def __init__(self) -> None:
        # Make the stage 0
        super().__init__(0)

    def summary_string(self) -> str:
        """ Returns a short string summarising the RowGenerator. """
        return "nothing"

    def _gen_row(self, previous_row: Row, stroke: Stroke, index: int) -> Row:
        raise NullRowGenError()



""" Helper functions for the row generation module. """

from typing import List, Optional

import itertools
import re

from wheatley.aliases import Places, Row
from wheatley.bell import Bell, BELL_NAMES

_CROSS_PN: Places = Places([])

STAGES = {
    "singles": 3,
    "minimus": 4,
    "doubles": 5,
    "minor": 6,
    "triples": 7,
    "major": 8,
    "caters": 9,
    "royal": 10,
    "cinques": 11,
    "maximus": 12,
    "sextuples": 13,
    "fourteen": 14,
    "septuples": 15,
    "sixteen": 16,
}


def valid_pn(pn_str: str) -> bool:
    """
    Check that a place notation string can be converted into a list of places.
    """
    if "," in pn_str:
        for part in pn_str.split(","):
            if not valid_pn(part):
                return False
        return True

    # Assumes a valid place notation string is delimited by `.`
    # These can optionally be omitted around an `-` or `x`
    # We substitute to ensure `-` is surrounded by `.` and replace any `..` caused by `--` => `.-..-.
    dot_delimited_string = re.sub("[.]*[x-][.]*", ".-.", pn_str).strip('.&+ ')
    deduplicated_string = dot_delimited_string.replace('..', '.').split('.')

    # Walk the places checking for either a bell crossing or a valid bell string
    for place in deduplicated_string:
        if place != '-':
            for y in place:
                if not y.upper() in BELL_NAMES:
                    return False

    return True


def convert_pn(pn_str: str, expect_symmetric: bool=False) -> List[Places]:
    """ Convert a place notation string into a list of places. """
    if "," in pn_str:
        return list(itertools.chain.from_iterable(convert_pn(part, True) for part in pn_str.split(",")))

    if expect_symmetric:
        symmetric = not pn_str.startswith('+')
    else:
        symmetric = pn_str.startswith('&')

    # Assumes a valid place notation string is delimited by `.`
    # These can optionally be omitted around an `-` or `x`
    # We substitute to ensure `-` is surrounded by `.` and replace any `..` caused by `--` => `.-..-.
    dot_delimited_string = re.sub("[.]*[x-][.]*", ".-.", pn_str).strip('.&+ ')
    deduplicated_string = dot_delimited_string.replace('..', '.').split('.')

    # We suppress the type error here, because mypy will assign the list comprehension type 'List[object]',
    # not 'List[Places]'.
    converted: List[Places] = [[convert_bell_string(y) for y in place] # type: ignore
                               if place != '-' else _CROSS_PN
                               for place in deduplicated_string]

    if symmetric:
        return converted + list(reversed(converted[:-1]))
    return converted


def convert_bell_string(bell: str) -> int:
    """ Convert a single-char string representing a bell into an integer. """
    try:
        return BELL_NAMES.index(bell) + 1
    except ValueError as e:
        raise ValueError(f"'{bell}' is not known bell symbol") from e


def convert_to_bell_string(bell: int) -> str:
    """ Convert an integer into the equivalent bell name. """
    if bell <= 0 or bell >= len(BELL_NAMES) + 1:
        raise ValueError(f"'{bell}' is not known bell number")

    return BELL_NAMES[bell - 1]


def rounds(number_of_bells: int) -> Row:
    """ Generate rounds on the given number of bells. """
    return Row([Bell.from_number(i) for i in range(1, number_of_bells + 1)])


def generate_starting_row(number_of_bells: int, start_row_string: Optional[str] = None) -> Row:
    """ Generate the starting row as rounds or a custom input. """
    if start_row_string is None:
        return rounds(number_of_bells)

    start_row = [Bell.from_str(i) for i in start_row_string]
    if len(start_row) > len(set(start_row)):
        raise ValueError(f"starting row '{start_row_string}' contains the same bell multiple times")

    # If there are more bells than given in the starting row
    # add the missing ones in sequential order as cover bells
    for i in range(1, number_of_bells + 1):
        if Bell.from_number(i) not in start_row:
            start_row.append(Bell.from_number(i))

    return Row(start_row)



""" A module to hold various classes determining when Wheatley rings his bells. """

from .wait_for_user import WaitForUserRhythm
from .regression import RegressionRhythm
from .abstract_rhythm import Rhythm



import unittest
from wheatley.stroke import Stroke, HANDSTROKE, BACKSTROKE


class StrokeTests(unittest.TestCase):
    def test_equality(self):
        for i in [True, False]:
            for j in [True, False]:
                # Strokes containing `i` and `j` should be equal precisely when `i == j`
                self.assertEqual(Stroke(i) == Stroke(j), i == j)
                # Strokes containing `i` and `j` should be not equal precisely when `i != j`
                self.assertEqual(Stroke(i) != Stroke(j), i != j)

    def test_constants(self):
        self.assertEqual(HANDSTROKE, Stroke(True))
        self.assertEqual(BACKSTROKE, Stroke(False))

    def test_is_hand(self):
        self.assertEqual(HANDSTROKE.is_hand(), True)
        self.assertEqual(BACKSTROKE.is_hand(), False)

    def test_is_back(self):
        self.assertEqual(HANDSTROKE.is_back(), False)
        self.assertEqual(BACKSTROKE.is_back(), True)

    def test_from_index(self):
        stroke = HANDSTROKE
        for i in range(-100, 100):
            self.assertEqual(Stroke.from_index(i), stroke)
            stroke = stroke.opposite()

    def test_opposite(self):
        self.assertEqual(HANDSTROKE.opposite(), BACKSTROKE)
        self.assertEqual(BACKSTROKE.opposite(), HANDSTROKE)

    def test_to_string(self):
        self.assertEqual(str(HANDSTROKE), 'HANDSTROKE')
        self.assertEqual(str(BACKSTROKE), 'BACKSTROKE')

    def test_to_char(self):
        self.assertEqual(HANDSTROKE.char(), 'H')
        self.assertEqual(BACKSTROKE.char(), 'B')


if __name__ == '__main__':
    unittest.main()



""" Module to handle generation of rows for Wheatley to ring. """

from .complib_composition_generator import ComplibCompositionGenerator
from .dixonoids_generator import DixonoidsGenerator
from .method_place_notation_generator import MethodPlaceNotationGenerator, generator_from_special_title
from .place_notation_generator import PlaceNotationGenerator
from .plain_hunt_generator import PlainHuntGenerator
from .row_generator import RowGenerator



import unittest
from unittest import TestCase


from wheatley.row_generation import ComplibCompositionGenerator


class CompLibGeneratorTests(TestCase):
    def test_comp_fetching(self):
        for (url, expected_title) in [
            ("complib.org/composition/62355", "5040 Plain Bob Major Op. Cyclic #1 by Ben White-Horne"),
            ("www.complib.org/composition/62355", "5040 Plain Bob Major Op. Cyclic #1 by Ben White-Horne"),
            ("https://www.complib.org/composition/71994", "110 2-Spliced Major"),
            ("https://complib.org/composition/71994", "110 2-Spliced Major"),
            (
                "https://complib.org/composition/70383?accessKey=cf75d5e0d213d4d3ea38f58f5bfdfd8e86b99ccf",
                "56 3-Spliced Bob Royal"
            )
        ]:
            with self.subTest(url=url, expected_title=expected_title):
                self.assertEqual(ComplibCompositionGenerator.from_arg(url).comp_title, expected_title)


if __name__ == '__main__':
    unittest.main()



""" A module to hold a row generator that produces plain hunt on any (even) stage. """

from wheatley.aliases import Places, Row
from wheatley.stroke import Stroke

from .row_generator import RowGenerator


class PlainHuntGenerator(RowGenerator):
    """ A row generator to create plain hunt on any stage. """

    def summary_string(self) -> str:
        """ Returns a short string summarising the RowGenerator. """
        return f"Plain Hunt on {self.stage}"

    def _gen_row(self, previous_row: Row, stroke: Stroke, index: int) -> Row:
        if stroke.is_hand():
            return self.permute(previous_row, Places([]))
        return self.permute(previous_row, Places([1, self.stage]))



""" Module to hold DixonoidsGenerator, a class for generating dixonoids. """

from typing import Dict, List, Optional

from wheatley.aliases import Places, Row
from wheatley.stroke import Stroke

from .helpers import convert_pn
from .row_generator import RowGenerator


class DixonoidsGenerator(RowGenerator):
    """ A class to generate rows of dixonoids. """

    DixonsRules = {
        0: ["x", "1"],
        1: ["x", "2"],
        2: ["x", "4"],
        4: ["x", "4"]
    }
    DefaultBob = {1: ["x", "4"]}
    DefaultSingle = {1: ["x", "1234"]}

    def __init__(self, stage: int, plain_rules: Dict[int, List[str]] = None,
                 bob_rules: Dict[int, List[str]] = None, single_rules: Dict[int, List[str]] = None,
                 start_row: Optional[str] = None) -> None:
        """
        Initialises a dixonoid generator.

        :param plain_rules: Dictionary of leading bell: [handstroke pn, backstroke pn]
                            0: Matches any other bell
        :param bob_rules: Dictionary of leading bell: [handstroke pn, backstroke pn]
                          Only include bells which lead when a bob is rung
        :param single_rules: Dictionary of leading bell: [handstroke pn, backstroke pn]
                          Only include bells which lead when a single is rung
        """
        super().__init__(stage, start_row)

        if plain_rules is None:
            plain_rules = self.DixonsRules

        if bob_rules is None:
            bob_rules = self.DefaultBob

        if single_rules is None:
            single_rules = self.DefaultSingle

        self.plain_rules = self._convert_pn_dict(plain_rules)
        self.bob_rules = self._convert_pn_dict(bob_rules)
        self.single_rules = self._convert_pn_dict(single_rules)

    def _gen_row(self, previous_row: Row, stroke: Stroke, index: int) -> Row:
        leading_bell = previous_row[0].number
        pn_index = 0 if stroke.is_hand() else 1

        if self._has_bob and self.bob_rules.get(leading_bell):
            place_notation = self.bob_rules[leading_bell][pn_index]

            if stroke.is_back():
                self.reset_calls()
        elif self._has_single and self.single_rules.get(leading_bell):
            place_notation = self.single_rules[leading_bell][pn_index]

            if stroke.is_back():
                self.reset_calls()
        elif self.plain_rules.get(leading_bell):
            place_notation = self.plain_rules[leading_bell][pn_index]
        else:
            place_notation = self.plain_rules[0][pn_index]

        row = self.permute(previous_row, place_notation)
        return row

    def summary_string(self) -> str:
        """ Returns a short string summarising the RowGenerator. """
        return f"dixonoid: {self.plain_rules}"

    @staticmethod
    def _convert_pn_dict(rules: Dict[int, List[str]]) -> Dict[int, List[Places]]:
        return {key: [convert_pn(pn)[0] for pn in places] for key, places in rules.items()}



import unittest

from wheatley.row_generation.helpers import convert_bell_string, convert_to_bell_string, \
    convert_pn, generate_starting_row
from wheatley.row_generation.helpers import _CROSS_PN
from wheatley.bell import MAX_BELL, Bell
from wheatley.aliases import Row

class HelpersBellStringTests(unittest.TestCase):
    def test_convert_bell_string(self):
        self.assertEqual(1, convert_bell_string("1"))
        self.assertEqual(9, convert_bell_string("9"))

        self.assertEqual(10, convert_bell_string("0"))
        self.assertEqual(11, convert_bell_string("E"))
        self.assertEqual(12, convert_bell_string("T"))
        self.assertEqual(13, convert_bell_string("A"))
        self.assertEqual(14, convert_bell_string("B"))
        self.assertEqual(15, convert_bell_string("C"))
        self.assertEqual(16, convert_bell_string("D"))

    def test_convert_bell_string_not_found(self):
        with self.assertRaisesRegex(ValueError, "'F' is not known bell symbol"):
            convert_bell_string("F")

    def test_convert_to_bell_string(self):
        self.assertEqual("1", convert_to_bell_string(1))
        self.assertEqual("9", convert_to_bell_string(9))

        self.assertEqual("0", convert_to_bell_string(10))
        self.assertEqual("E", convert_to_bell_string(11))
        self.assertEqual("T", convert_to_bell_string(12))
        self.assertEqual("A", convert_to_bell_string(13))
        self.assertEqual("B", convert_to_bell_string(14))
        self.assertEqual("C", convert_to_bell_string(15))
        self.assertEqual("D", convert_to_bell_string(16))

    def test_convert_to_bell_string_not_found(self):
        with self.assertRaisesRegex(ValueError, "'0' is not known bell number"):
            convert_to_bell_string(0)

        with self.assertRaisesRegex(ValueError, f"'{MAX_BELL + 1}' is not known bell number"):
            convert_to_bell_string(MAX_BELL + 1)


class HelpersPlaceNotationTests(unittest.TestCase):
    def test_convert_pn_unknown_bell(self):
        with self.assertRaisesRegex(ValueError, "'F' is not known bell symbol"):
            place_notation = "&-F"
            convert_pn(place_notation)

    def test_convert_pn(self):
        test_cases = [
            ("&-1", [_CROSS_PN, [1], _CROSS_PN]),
            ("-1", [_CROSS_PN, [1]]),
            ("-123-4", [_CROSS_PN, [1, 2, 3], _CROSS_PN, [4]]),
            ("-1-1,2", [_CROSS_PN, [1], _CROSS_PN, [1], _CROSS_PN, [1], _CROSS_PN, [2]]),
            ("x1", [_CROSS_PN, [1]]),
            ("x.1.-.", [_CROSS_PN, [1], _CROSS_PN]),
            ("12.3-123", [[1, 2], [3], _CROSS_PN, [1, 2, 3]]),
            ("&-1,&2.3", [_CROSS_PN, [1], _CROSS_PN, [2], [3], [2]]),
            ("&-1,2.3", [_CROSS_PN, [1], _CROSS_PN, [2], [3], [2]]),
            ("&-1,+2.3", [_CROSS_PN, [1], _CROSS_PN, [2], [3]]),
            ("-1,&2.3", [_CROSS_PN, [1], _CROSS_PN, [2], [3], [2]]),
            ("-1,2.3", [_CROSS_PN, [1], _CROSS_PN, [2], [3], [2]]),
            ("-1,2.3,4", [_CROSS_PN, [1], _CROSS_PN, [2], [3], [2], [4]]),
            ("---4", [_CROSS_PN, _CROSS_PN, _CROSS_PN, [4]]),
            (".-.-.1.-.2", [_CROSS_PN, _CROSS_PN, [1], _CROSS_PN, [2]])
        ]

        for (input_pn, expected_output_pn) in test_cases:
            with self.subTest(input_pn=input_pn, expected_output_pn=expected_output_pn):
                self.assertEqual(convert_pn(input_pn), expected_output_pn)

class HelpersGenerateStartingRowTests(unittest.TestCase):
    def test_generate_without_custom_row(self):
        self.assertListEqual(Row([Bell(0), Bell(1), Bell(2), Bell(3)]), generate_starting_row(4))

    def test_generate_with_custom_row(self):
        self.assertListEqual(Row([Bell(3), Bell(2), Bell(1), Bell(0)]), generate_starting_row(4, "4321"))
        self.assertListEqual(Row([Bell(3), Bell(2), Bell(1), Bell(0)]), generate_starting_row(4, "432"))
        self.assertListEqual(Row([Bell(3), Bell(1), Bell(0), Bell(2)]), generate_starting_row(4, "42"))
        self.assertListEqual(Row([Bell(1), Bell(0), Bell(2), Bell(3)]), generate_starting_row(4, "2"))

    def test_generate_with_invalid_custom_row(self):
        with self.assertRaisesRegex(ValueError, "starting row '4324' contains the same bell multiple times"):
            generate_starting_row(4, "4324")

if __name__ == '__main__':
    unittest.main()



""" A module containing a type that encapsulates a stroke. """

from typing import Any


class Stroke:
    """ A new-type of 'bool' that encapsulates a stroke: i.e. HANDSTROKE or BACKSTROKE. """

    def __init__(self, is_handstroke: bool) -> None:
        self._is_handstroke = is_handstroke

    def is_hand(self) -> bool:
        """ Returns true if this Stroke represents a handstroke.
        Equivalent to `stroke == HANDSTROKE`.
        """
        return self._is_handstroke

    def is_back(self) -> bool:
        """ Returns true if this Stroke represents a backstroke.
        Equivalent to `stroke == BACKSTROKE`.
        """
        return not self._is_handstroke

    @classmethod
    def from_index(cls, index: int) -> 'Stroke':
        """ Returns the stroke of the row that would exist at a given index. """
        return cls(index % 2 == 0)

    def opposite(self) -> 'Stroke':
        """ Returns the opposite Stroke to the current one. """
        return Stroke(not self._is_handstroke)

    def char(self) -> str:
        """ Returns a single-character string of 'H' or 'B'. """
        return "H" if self._is_handstroke else "B"

    def __str__(self) -> str:
        return "HANDSTROKE" if self._is_handstroke else "BACKSTROKE"

    def __repr__(self) -> str:
        return str(self)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Stroke):
            return other._is_handstroke == self._is_handstroke
        return False

    def __ne__(self, other: Any) -> bool:
        return not self == other

    def __inverse__(self) -> 'Stroke':
        return self.opposite()

    def __hash__(self) -> int:
        return self._is_handstroke.__hash__()


HANDSTROKE: Stroke = Stroke(True)
BACKSTROKE: Stroke = Stroke(False)



"""
A module to contain all the functions that parse command line arguments, and their error classes.
"""

from typing import Dict, NoReturn, Optional, Tuple
import logging

from wheatley.aliases import CallDef, JSON
from wheatley.bell import Bell
from wheatley.row_generation import RowGenerator
from wheatley.row_generation.place_notation_generator import PlaceNotationGenerator
from wheatley.row_generation.complib_composition_generator import ComplibCompositionGenerator, \
                                                                  PrivateCompError, InvalidCompError
from wheatley.row_generation.helpers import valid_pn

class StartRowParseError(ValueError):
    """
    An error thrown with a helpful error message when the user inputs a start row string that
    cannot be parsed by the parser.
    """

    def __init__(self, start_row_string: str, message: str) -> None:
        super().__init__()

        self.start_row_string = start_row_string
        self.message = message

    def __str__(self) -> str:
        return f"Error parsing start row '{self.start_row_string}': {self.message}"

def parse_start_row(start_row: str) -> int:
    """
    Parses a start row written in the format "12345" to ensure all characters are valid bells
    and there are no missing bells.
    """
    def exit_with_message(error_text: str) -> NoReturn:
        """ Raise an exception with a useful error message. """
        raise StartRowParseError(start_row, error_text)

    if start_row is None:
        return 0

    max_bell: int = 0

    if start_row is None:
        return max_bell

    # First loop checks for invalid bells and finds the largest bell in the row
    for char in start_row:
        try:
            bell = Bell.from_str(char)
        except ValueError as e:
            exit_with_message(str(e))

        max_bell = max(max_bell, bell.number)

    # Second loop checks if there are duplicates and if every bell is specified
    expected_bells = list(range(1, max_bell + 1))
    for char in start_row:
        bell_number = Bell.from_str(char).number

        if bell_number not in expected_bells:
            exit_with_message(f"Start row contains bell {bell_number} mutiple times")

        expected_bells.remove(bell_number)

    if len(expected_bells) > 0:
        exit_with_message(f"Start row does not contain bell(s) {expected_bells}")

    return len(start_row)

class PealSpeedParseError(ValueError):
    """
    An error thrown with a helpful error message when the user inputs a peal speed string that
    cannot be parsed by the parser.
    """

    def __init__(self, peal_speed_string: str, message: str) -> None:
        super().__init__()

        self.peal_speed_string = peal_speed_string
        self.message = message

    def __str__(self) -> str:
        return f"Error parsing peal speed '{self.peal_speed_string}': {self.message}"


def parse_peal_speed(peal_speed: str) -> int:
    """
    Parses a peal speed written in the format /2h58(m?)/ or /XXX(m?)/ into a number of minutes.
    """
    def exit_with_message(error_text: str) -> NoReturn:
        """ Raise an exception with a useful error message. """
        raise PealSpeedParseError(peal_speed, error_text)

    # Strip whitespace from the argument, so that if the user is in fact insane enough to pad their
    # CLI arguments with whitespace then they can do so and not crash the program.  This also has
    # the side effect of cloning the input string so we can freely modify it.
    stripped_peal_speed = peal_speed.strip()

    # Remove the 'm' from the end of the peal speed - it doesn't add any clarity
    if stripped_peal_speed.endswith("m"):
        stripped_peal_speed = stripped_peal_speed[:-1]

    if "h" in stripped_peal_speed:
        # Split the peal speed into its hour and minute components, and print a helpful message
        # if there are too many parts
        split_parts = stripped_peal_speed.split("h")
        if len(split_parts) > 2:
            exit_with_message("The peal speed should contain at most one 'h'.")

        hour_string, minute_string = split_parts
        # Strip the input values so that the user can put whitespace into the input if they want
        hour_string = hour_string.strip()
        minute_string = minute_string.strip()

        # Parse the hours value, and print messages if it is invalid
        try:
            hours = int(hour_string)
        except ValueError:
            exit_with_message(f"The hour value '{hour_string}' is not an integer.")

        if hours < 0:
            exit_with_message(f"The hour value '{hour_string}' must be a positive integer.")

        # Parse the minute value, and print messages if it is invalid
        try:
            minutes = 0 if minute_string == "" else int(minute_string)
        except ValueError:
            exit_with_message(f"The minute value '{minute_string}' is not an integer.")

        if minutes < 0:
            exit_with_message(f"The minute value '{minute_string}' must be a positive integer.")
        if minutes > 59:
            exit_with_message(f"The minute value '{minute_string}' must be smaller than 60.")

        return hours * 60 + minutes

    # If the user doesn't put an 'h' in their string, then we assume it's just a minute value that
    # may or may not be bigger than 60
    try:
        minutes = int(stripped_peal_speed)
    except ValueError:
        exit_with_message(f"The minute value '{stripped_peal_speed}' is not an integer.")

    if minutes < 0:
        exit_with_message(f"The minute value '{stripped_peal_speed}' must be a positive integer.")

    return minutes


class CallParseError(ValueError):
    """
    An error thrown with a helpful error message when the user inputs a peal speed string that
    cannot be parsed by the parser.
    """

    def __init__(self, call_string: str, message: str) -> None:
        super().__init__()

        self.call_string = call_string
        self.message = message

    def __str__(self) -> str:
        return f"Error parsing call string '{self.call_string}': {self.message}"


def parse_call(input_string: str) -> CallDef:
    """ Parse a call string into a dict of lead locations to place notation strings. """
    def exit_with_message(message: str) -> NoReturn:
        """ Raises a parse error with the given error message. """
        raise CallParseError(input_string, message)

    # A dictionary that will be filled with the parsed calls
    parsed_calls: Dict[int, str] = {}

    for segment in input_string.split("/"):
        # Default the location to 0 and initialise place_notation_str with None
        location = 0
        place_notation_str = None

        if ":" in segment:
            # Split the segment into location and place notations
            try:
                location_str, place_notation_str = segment.split(":")
            except ValueError:
                exit_with_message(
                    f"Call specification '{segment.strip()}' should contain at most one ':'."
                )

            # Strip whitespace from the string segments so that they can be parsed more easily
            assert place_notation_str is not None
            location_str = location_str.strip()
            place_notation_str = place_notation_str.strip()

            # Parse the first section as an integer
            try:
                location = int(location_str)
            except ValueError:
                exit_with_message(f"Location '{location_str}' is not an integer.")
        else:
            # If there's only one section, it must be the place notation so all we need to do is to
            # strip it of whitespace (and `location` defaults to 0 so that's also set correctly)
            place_notation_str = segment.strip()

        # Produce a nice error message if the pn string is empty
        if place_notation_str == "":
            exit_with_message("Place notation strings cannot be empty.")

        # Insert the new call definition into the dictionary
        if location in parsed_calls:
            exit_with_message(f"Location {location} has two conflicting calls: \
'{parsed_calls[location]}' and '{place_notation_str}'.")

        parsed_calls[location] = place_notation_str

    return CallDef(parsed_calls)


class RowGenParseError(ValueError):
    """ A class to encapsulate an error generated when parsing an RowGenerator from JSON. """
    def __init__(self, json: JSON, field: str, message: str) -> None:
        super().__init__()

        self.json = json
        self.field = field
        self.message = message

    def __str__(self) -> str:
        return f"Error parsing RowGen json '{self.json}' in field '{self.field}': {self.message}"

    def __repr__(self) -> str:
        return str(self)


def json_to_row_generator(json: JSON, logger: logging.Logger) -> RowGenerator:
    """ Takes a JSON message from SocketIO and convert it to a RowGenerator or throw an exception. """
    def raise_error(field: str, message: str, parent_error: Optional[Exception] = None) -> NoReturn:
        """ A helper function to raise a `RowGenParseError` with a helpful error message. """
        if parent_error is None:
            raise RowGenParseError(json, field, message)
        raise RowGenParseError(json, field, message) from parent_error

    def json_to_call(name: str) -> Optional[CallDef]:
        """ Helper function to generate a call with a given name from the json. """
        if name not in json:
            logger.warning(f"No field '{name}' in the row generator JSON")
            return None

        call: Dict[int, str] = {}
        for key, value in json[name].items():
            try:
                index = int(key)
            except ValueError as e:
                raise_error(name, f"Call index '{key}' is not a valid integer", e)
            call[index] = value
        return CallDef(call)

    if 'type' not in json:
        raise_error('type', "'type' is not defined")

    if json['type'] == 'method':
        try:
            stage = int(json['stage'])
        except KeyError as e:
            raise_error('stage', "'stage' is not defined", e)
        except ValueError as e:
            raise_error('stage', f"'{json['stage']}' is not a valid integer", e)
        return PlaceNotationGenerator(stage, json['notation'], json_to_call('bob'),
                                      json_to_call('single'))

    if json['type'] == "composition":
        try:
            comp_url = json['url']
        except KeyError as e:
            raise_error('url', "'url' is not defined", e)

        try:
            row_gen = ComplibCompositionGenerator.from_arg(comp_url)
        except PrivateCompError as e:
            raise_error('complib request', "Comp id '{comp_url}' is private", e)
        except InvalidCompError as e:
            raise_error('complib request', "No composition with id '{comp_url}' found", e)
        return row_gen

    raise_error('type', f"{json['type']} is not one of 'method' or 'composition'")


def to_bool(value: str) -> bool:
    """
    Converts a string argument to the bool representation (or throws a ValueError if the value is not
    one of '[Tt]rue' or '[Ff]alse'.
    """
    if value in ["True", "true", True]:
        return True
    if value in ["False", "false", False]:
        return False
    raise ValueError(f"Value {value} cannot be converted into a bool")


class PlaceNotationError(ValueError):
    """
    An error thrown with a helpful error message when the user inputs a place
    notation string that cannot be parsed.
    """

    def __init__(self, place_notation: str, message: str) -> None:
        super().__init__()

        self.place_notation = place_notation
        self.message = message

    def __str__(self) -> str:
        return f"Error in place notation '{self.place_notation}': {self.message}"


def parse_place_notation(input_string: str) -> Tuple[int, str]:
    """
    Parse a stage and place notation throwing an error when an invalid stage
    or place notation specifier is found
    """

    # Looking for a string that matches <stage>:<place notation> where the
    # place notation is a series of bell numbers and 'x' characters
    parts = input_string.split(':')
    if len(parts) == 2:
        stage_part = parts[0]
        if len(stage_part) == 0 or not stage_part.isnumeric():
            raise PlaceNotationError(input_string, 'Stage must be a number')
        stage = int(stage_part)
        place_notation = parts[1]
        if not valid_pn(place_notation):
            raise PlaceNotationError(input_string, 'Place notation is invalid')
    else:
        raise PlaceNotationError(input_string, '<stage>:<place notation> required')

    return stage, place_notation



from typing import List, Iterator

from wheatley.bell import Bell
from wheatley.stroke import Stroke, HANDSTROKE


def rounds(stage: int) -> List[int]:
    number_of_bells = stage + 1 if stage % 2 else stage
    return [i for i in range(1, number_of_bells + 1)]


def as_bells(nums: List[int]) -> List[Bell]:
    return [Bell.from_number(x) for x in nums]


def gen_rows(generator, count, start_stroke: Stroke = HANDSTROKE) -> List[int]:
    return list(yield_rows(generator, count, start_stroke))


def yield_rows(generator, count, start_stroke=HANDSTROKE) -> Iterator[int]:
    stroke = start_stroke
    for _ in range(count):
        yield [bell.number for bell in generator.next_row(stroke)]
        stroke = stroke.opposite()



""" Utils for the fuzzing code. """

class FuzzingError(ValueError):
    """ A custom error used to signal that some un-abstracted errors were found. """

    def __init__(self, function_name, errors_found):
        super().__init__()

        self._function_name = function_name
        self._errors_found = errors_found

    def __str__(self):
        error_messages = "\n  ".join(self._errors_found)

        return f"'{self._function_name}' threw uncaught errors:\n{error_messages}"


def fuzz_for_unwrapped_errors(function_name,
                              function_to_fuzz,
                              input_generation_function,
                              expected_error,
                              iterations=100000):
    """ Fuzz a given function with generated input, checking for a given error. """

    errors_found = []

    for _ in range(iterations):
        generated_input = input_generation_function()

        try:
            function_to_fuzz(generated_input)
        except expected_error:
            pass
        except Exception as e:
            error_type = str(type(e))

            if "'" in error_type:
                error_type = error_type.split("'")[1]

            errors_found.append(
                f"Error type '{error_type}' thrown on input '{generated_input}': '{e}'."
            )

            if len(errors_found) > 10:
                break

    if len(errors_found) == 0:
        print(f"{iterations} iterations were run on {function_name}, and they all threw the right \
error class.")
    else:
        raise FuzzingError(function_name, errors_found)



""" Run a fuzzer to generate lots of random input for the call parser. """

from random import choice, randint

from wheatley.parsing import parse_call, CallParseError

from .fuzz_utils import fuzz_for_unwrapped_errors


def random_call_string():
    """ Generate a plausible input value for `wheatley.arg_parsing.parse_call` to parse. """

    return "".join([choice("1234567890ET:/.&#xx#?s ") for _ in range(randint(0, 20))])


def fuzz_parse_call():
    fuzz_for_unwrapped_errors("parse_call", parse_call, random_call_string, CallParseError)



""" Module for the base Rhythm class, from which all other Rhythm classes inherit. """

import time

from abc import ABCMeta, abstractmethod
from typing import Any

from wheatley.stroke import Stroke
from wheatley.bell import Bell


class Rhythm(metaclass=ABCMeta):
    """
    An abstract Rhythm class, used as an interface by the Bot class to interact with Rhythms.
    """

    @abstractmethod
    def return_to_mainloop(self) -> None:
        """
        When called, this function requires that Wheatley's main thread should return from
        wait_for_bell_time and give control back to the mainloop within a reasonable amount of time
        (i.e. 1/10 of a second is acceptable, but anything approacing a second is not)
        """

    @abstractmethod
    def wait_for_bell_time(self, current_time: float, bell: Bell, row_number: int, place: int,
                           user_controlled: bool, stroke: Stroke) -> None:
        """ Sleeps the thread until a given Bell should have rung. """

    @abstractmethod
    def expect_bell(self, expected_bell: Bell, row_number: int, place: int, expected_stroke: Stroke) -> None:
        """
        Indicates that a given Bell is expected to be rung at a given row, place and stroke.
        Used by the rhythm so that when that bell is rung later, it can tell where that bell
        _should_ have been in the ringing, and so can use that knowledge to inform the speed of the
        ringing.
        """

    @abstractmethod
    def change_setting(self, key: str, value: Any, real_time: float) -> None:
        """ Called when the Ringing Room server asks Wheatley to change a setting. """
        # We need the `real_time` param because `WaitForUserRhythm` lies to its inner rhythm about
        # the real time (in order to hold up over people without killing the regression), and
        # `RegressionRhythm` needs the current 'time' so that it can update its line accurately when
        # the peal speed is changed.

    @abstractmethod
    def on_bell_ring(self, bell: Bell, stroke: Stroke, real_time: float) -> None:
        """
        Called when a bell is rung at a given stroke.  Used as a callback from the Tower class.
        """

    @abstractmethod
    def initialise_line(self, stage: int, user_controls_treble: bool, start_time: float,
                        number_of_user_controlled_bells: int) -> None:
        """ Allow the Rhythm object to initialise itself when 'Look to' is called. """

    def sleep(self, seconds: float) -> None: #  pylint: disable=no-self-use
        """ Sleeps for given number of seconds. Allows mocking in tests"""
        time.sleep(seconds)



""" Module to hold a RowGenerator to generate rows from a method title. """

import xml.etree.ElementTree as ET
from typing import Optional, Tuple

import requests

from wheatley.aliases import CallDef

from .dixonoids_generator import DixonoidsGenerator
from .helpers import STAGES
from .plain_hunt_generator import PlainHuntGenerator
from .place_notation_generator import PlaceNotationGenerator
from .row_generator import RowGenerator


def generator_from_special_title(method_title: str, start_row: Optional[str] = None) \
    -> Optional[RowGenerator]:
    """ Creates a row generator from special method titles. """
    lowered_title = method_title.lower().strip()
    if " " not in lowered_title:
        raise MethodNotFoundError(method_title)

    method_name, stage_name = lowered_title.rsplit(" ", 1)
    method_name = method_name.strip()

    if stage_name.isdigit() and int(stage_name) in STAGES.values():
        stage = int(stage_name)
    elif stage_name in STAGES:
        stage = STAGES[stage_name]
    else:
        raise MethodNotFoundError(method_title)

    if method_name == "grandsire" and stage >= 5:
        return PlaceNotationGenerator.grandsire(stage, start_row)
    if method_name == "stedman" and stage % 2 and stage >= 5:
        return PlaceNotationGenerator.stedman(stage, start_row)
    if method_name in ["plain hunt", "plain hunt on"]:
        return PlainHuntGenerator(stage, start_row)
    if method_name == "dixon's bob" and stage == 6:
        return DixonoidsGenerator(stage, start_row=start_row)
    return None


class MethodNotFoundError(ValueError):
    """
    An error class to store the error thrown when a method title is inputted that doesn't have an
    entry in the CC method library.
    """

    def __init__(self, name: str) -> None:
        super().__init__()

        self._name = name

    def __str__(self) -> str:
        return f"No method with title '{self._name}' found."


class MethodPlaceNotationGenerator(PlaceNotationGenerator):
    """ A class to generate rows given a method title. """

    def __init__(self, method_title: str, bob: CallDef, single: CallDef,
                 start_row: str, start_index: int = 0) -> None:
        method_xml = self._fetch_method(method_title)

        try:
            method_pn, stage, actual_title = self._parse_xml(method_xml)
            self._title = actual_title
        except AttributeError as e:
            raise MethodNotFoundError(method_title) from e

        super().__init__(
            stage,
            method_pn,
            bob,
            single,
            start_index,
            start_row
        )

    def summary_string(self) -> str:
        """ Returns a short string summarising the RowGenerator. """
        return f"{self._title}"

    @staticmethod
    def _parse_xml(method_xml: str) -> Tuple[str, int, str]:
        method_parsed_xml = ET.fromstring(method_xml)
        xmlns = '{http://methods.ringing.org/NS/method}'

        # Unpack the title
        title_elems = method_parsed_xml.findall(xmlns + 'method/' + xmlns + 'title')
        if len(title_elems) == 0:
            raise AttributeError("No method title found in XML")
        title: Optional[str] = title_elems[0].text
        assert title is not None

        # Schema at http://methods.ringing.org/xml.html
        symblock = method_parsed_xml.findall(xmlns + 'method/' + xmlns + 'pn/' + xmlns + 'symblock')
        block = method_parsed_xml.findall(xmlns + 'method/' + xmlns + 'pn/' + xmlns + 'block')
        # Unpack the stage piece-by-piece to appease the type checker and assert expected types at runtime
        maybe_stage_str = method_parsed_xml.find(xmlns + 'method/' + xmlns + 'stage')
        assert maybe_stage_str is not None
        stage_str = maybe_stage_str.text
        assert isinstance(stage_str, str)
        stage = int(stage_str)

        if len(symblock) != 0:
            notation = symblock[0].text
            lead_end = symblock[1].text
            return f"&{notation},&{lead_end}", stage, title

        if len(block) != 0:
            notation = block[0].text
            assert notation is not None
            return notation, stage, title

        raise Exception("Place notation not found")

    @staticmethod
    def _fetch_method(method_title: str) -> str:
        params = {'title': method_title, 'fields': 'title|pn|stage'}
        source = requests.get('http://methods.ringing.org/cgi-bin/simple.pl', params=params)
        source.raise_for_status()

        return source.text



""" A module for the Rhythm class that detects peal speed by performing regression on the users' bells. """

import logging
import math

from typing import Any, Dict, List, Tuple

# We can disable the type checking on numpy because it is only used within this file and the interface from
# this file to the rest of the code is type-checked.
import numpy # type: ignore

from wheatley.stroke import Stroke
from wheatley.bell import Bell
from .abstract_rhythm import Rhythm


WEIGHT_REJECTION_THRESHOLD = 0.001


# ===== REGRESSION FUNCTIONS =====

def fill(index: int, item: float, length: int) -> List[float]:
    """
    Make an array that contains `length` 0s, but with the value at `index` replaced with `item`.
    """
    a = [0.0 for _ in range(length)]
    a[index] = item
    return a


def calculate_regression(data_set: List[Tuple[float, float, float]]) -> Tuple[float, float]:
    """
    Calculates a weighted linear regression over the data given in data_set.
    Expects data_set to consist of 3-tuples of (blow_time, real_time, weight).
    """
    blow_times = [b for (b, r, w) in data_set]
    real_times = [r for (b, r, w) in data_set]
    weights = [w for (b, r, w) in data_set]

    num_datapoints = len(weights)

    x = numpy.array([[1, b] for b in blow_times])
    w = numpy.array([fill(i, w, num_datapoints) for (i, w) in enumerate(weights)])
    y = numpy.array([[x] for x in real_times])

    # Calculate (X^T * W * X)^-1 * (X^T * W * y)
    beta = numpy.linalg.inv(x.transpose().dot(w).dot(x)).dot(x.transpose()).dot(w).dot(y)

    return beta[0][0], beta[1][0]



# ===== UTILITY FUNCTIONS =====

def peal_speed_to_blow_interval(peal_minutes: float, num_bells: int) -> float:
    """ Calculate the blow interval from the peal speed, assuming a peal of 5040 changes """
    peal_speed_seconds = peal_minutes * 60
    seconds_per_whole_pull = peal_speed_seconds / 2520  # 2520 whole pulls = 5040 rows
    return seconds_per_whole_pull / (num_bells * 2 + 1)


def inverse_lerp(a: float, b: float, c: float) -> float:
    """
    Inverse function to `lerp`.  Calculates t such that lerp(a, b, t) = c.
    (will divide by zero if a = b)
    """
    return (c - a) / (b - a)


def lerp(a: float, b: float, t: float) -> float:
    """
    Linearly interpolates (unclamped) between a and b with factor t.
    Acts such that `lerp(a, b, 0.0)` returns `a`, and `lerp(a, b, 1.0)` returns `b`.
    """
    return (1 - t) * a + t * b


# ===== RHYTHM CLASS =====

class RegressionRhythm(Rhythm):
    """
    A class that will use regression to figure out the current ringing speed and ring accordingly.
    """

    logger_name = "RHYTHM:Regression"

    def __init__(self, inertia: float, peal_speed: float = 178, handstroke_gap: float = 1,
                 min_bells_in_dataset: int = 4, max_bells_in_dataset: int = 15,
                 initial_inertia: float = 0) -> None:
        """
        Initialises a new RegressionRhythm with a given default peal speed and handstroke gap.
        `peal_speed` is the number of minutes in a peal of 5040 changes, and `handstroke_gap`
        measures how many places long the handstroke gap is.  `initial_inertia` sets the inertia for only
        the first row, to aid with a smooth pulloff.
        """
        # An inertia-like coefficient designed to allow the regression finder to slowly adjust to
        # a new rhythm
        # 0.0 means that a new regression line will take effect instantly
        # 1.0 means that no effect is made at all
        self._preferred_inertia = inertia
        self._initial_inertia = initial_inertia

        self._peal_speed = peal_speed
        self._handstroke_gap = handstroke_gap
        self._min_bells_in_dataset = min_bells_in_dataset
        self._max_bells_in_dataset = max_bells_in_dataset

        self.stage = 0
        # Two values that determine the linear relationship between ringing and real time - these are set by
        # the regression algorithm and used to decide when Wheatley's bells should be rung
        self._start_time = 0.0
        self._blow_interval = 0.0

        # The actual time that the current touch started (this is unaffected by the regression algorithm)
        self._real_start_time = 0.0

        self._number_of_user_controlled_bells = 0
        # Maps a bell and a stroke to the row number and place which that bell is next expected to ring at
        # that stroke
        self._expected_bells: Dict[Tuple[Bell, Stroke], Tuple[int, int]] = {}
        self.data_set: List[Tuple[float, float, float]] = []

        self.logger = logging.getLogger(self.logger_name)

        self._should_return_to_mainloop = False

    def _add_data_point(self, row_number: int, place: int, real_time: float, weight: float) -> None:
        # Always manage the data set, even when inertia is 1 (in case inertia gets changed whilst
        # Wheatley is ringing)
        blow_time = self.index_to_blow_time(row_number, place)
        self.data_set.append((blow_time, real_time, weight))
        for (b, r, w) in self.data_set:
            self.logger.debug(f"Datapoint: {b:4.1f} {r - self._real_start_time:8.3f}s {w:.3f}")
        # Filter out datapoints with extremely low weights
        self.data_set = list(filter(lambda d: d[2] > WEIGHT_REJECTION_THRESHOLD, self.data_set))
        # Eventually forget about datapoints
        if len(self.data_set) >= self._max_bells_in_dataset:
            del self.data_set[0]

        # The inertia is set to 0 for the first row, to make sure that there's a smooth pullof
        inertia = self._preferred_inertia if row_number > 0 else self._initial_inertia
        # Early return if inertia == 1, since there's no point doing any regression if the result
        # will be ignored
        if inertia == 1:
            return

        # Only calculate the regression line if there are enough datapoints to gain a meaningful result
        if len(self.data_set) >= self._min_bells_in_dataset:
            (new_start_time, new_blow_interval) = calculate_regression(self.data_set)
            # Lerp between the new times and the old times, according to the desired inertia
            self._start_time = lerp(new_start_time, self._start_time, inertia)
            self._blow_interval = lerp(new_blow_interval, self._blow_interval, inertia)
            self.logger.debug(f"Start time: {self._start_time:.3f}")
            self.logger.debug(f"Bell interval: {self._blow_interval:.3f}")

    def return_to_mainloop(self) -> None:
        """
        When called, this function requires that Wheatley's main thread should return from
        wait_for_bell_time and give control back to the mainloop within a reasonable amount of time
        (i.e. 1/10 of a second is acceptable, but anything approacing a second is not)
        """
        self._should_return_to_mainloop = True

    def wait_for_bell_time(self, current_time: float, bell: Bell, row_number: int, place: int,
                           user_controlled: bool, stroke: Stroke) -> None:
        """ Sleeps the thread until a given Bell should have rung. """
        if user_controlled and self._start_time == float('inf'):
            self.logger.debug("Waiting for pull off")

            while self._start_time == float('inf'):
                self.sleep(0.01)

            self.logger.debug("Pulled off")

            return

        bell_time = self.index_to_real_time(row_number, place)
        if bell_time == float('inf') or self._start_time == 0:
            self.logger.error(f"Bell Time {bell_time}; Start Time {self._start_time}")
            self.sleep(self._blow_interval or 0.2)
        elif bell_time > current_time:
            self.sleep(bell_time - current_time)
        else:
            # Slow the ticks slightly
            self.sleep(0.01)

        self._should_return_to_mainloop = False

    def expect_bell(self, expected_bell: Bell, row_number: int, place: int, expected_stroke: Stroke) -> None:
        """
        Indicates that a given Bell is expected to be rung at a given row, place and stroke.
        Used by the rhythm so that when that bell is rung later, it can tell where that bell
        _should_ have been in the ringing, and so can use that knowledge to inform the speed of the
        ringing.
        """
        self.logger.debug(f"Expected bell {expected_bell} at index {row_number}:{place} at {expected_stroke}")
        self._expected_bells[(expected_bell, expected_stroke)] = (row_number, place)

    def change_setting(self, key: str, value: Any, real_time: float) -> None:
        def log_warning(message: str) -> None:
            self.logger.warning(f"Invalid value for setting '{key}': {message}")

        if key == "sensitivity":
            self.logger.warning(f"NOT IMPLEMENTED: setting sensitivity to {value}.")
        if key == "inertia":
            try:
                new_inertia = float(value)
                if new_inertia > 1 or new_inertia < 0:
                    log_warning(f"{new_inertia} is not between 0 and 1")
                else:
                    self._preferred_inertia = new_inertia
                    self.logger.warning(f"Setting 'self._preferred_inertia' to '{value}'")
            except ValueError:
                log_warning(f"{value} is not an number")
        if key == "peal_speed":
            new_peal_speed = int(value)
            if new_peal_speed <= 0:
                log_warning(f"{new_peal_speed} is not positive")
            else:
                self._peal_speed = new_peal_speed
                # Don't try to calculate the regression line if _blow_interval is 0 (because the lin
                # alg will cause a DivisionByZeroError).
                if self._blow_interval == 0:
                    return
                # Make sure that the `Peal Speed` control on Ringing Room can update the speed in
                # real time - this change makes no difference to the CLI version, because
                # `self._peal_speed` cannot be changed whilst Wheatley is running.  This is not a
                # long term solution (famous last words ;P), but for the initial RR release, I think
                # this should happen.
                peal_speed_interval = peal_speed_to_blow_interval(self._peal_speed, self.stage)
                current_bell_time = self.real_time_to_blow_time(real_time)
                # We want to find the line which has gradient `peal_speed_interval` and intersects with
                # our current line at the current time (so that Wheatley's perception of time doesn't
                # simply jump).
                self._blow_interval = peal_speed_interval
                self._start_time = real_time - current_bell_time * self._blow_interval
                self.logger.info(
                    f"Changing peal speed to {self._peal_speed} (bell interval {self._blow_interval:.3})"
                )

    def on_bell_ring(self, bell: Bell, stroke: Stroke, real_time: float) -> None:
        """
        Called when a bell is rung at a given stroke.  Used as a callback from the Tower class.
        """
        # If this bell was expected at this stroke (i.e. is being rung by someone else)
        if (bell, stroke) in self._expected_bells:
            # Figure out where the bell was expected in ringing space
            (row_number, place) = self._expected_bells[(bell, stroke)]
            expected_blow_time = self.index_to_blow_time(row_number, place)
            diff = self.real_time_to_blow_time(real_time) - expected_blow_time

            self.logger.debug(f"{bell} off by {diff:.3f} places")

            # If this was the first bell, then overwrite the start_time to update
            # the regression line
            if expected_blow_time == 0:
                self._start_time = real_time

            # Calculate the weight (which will be 1 if it is either of the first two bells to be
            # rung to not skew the data from the start)
            weight = math.exp(- diff ** 2)
            if len(self.data_set) <= 1:
                weight = 1

            # Add the bell as a datapoint with the calculated weight
            self._add_data_point(
                row_number,
                place,
                real_time,
                weight
            )

            del self._expected_bells[(bell, stroke)]
        else:
            # If this bell wasn't expected, then log that
            self.logger.warning(f"Bell {bell} unexpectedly rang at {stroke}")

    def initialise_line(self, stage: int, user_controls_treble: bool, start_time: float,
                        number_of_user_controlled_bells: int) -> None:
        """ Allow the Rhythm object to initialise itself when 'Look to' is called. """
        self._number_of_user_controlled_bells = number_of_user_controlled_bells
        self.stage = stage

        # Remove any data that's left over in the dataset from the last touch
        self.data_set = []

        # Calculate the blow interval from the peal speed, asuming a peal of 5040 changes
        self._blow_interval = peal_speed_to_blow_interval(self._peal_speed, self.stage)

        # Set real start time to make debug output more succinct
        self._real_start_time = start_time

        if not user_controls_treble:
            # If Wheatley is ringing the first bell, then add it as a datapoint anyway, so that after
            # the 2nd bell is rung, a regression line can be made
            self._add_data_point(0, 0, start_time, 1)
            self._start_time = start_time
        else:
            # If Wheatley isn't ringing the first bell, then set the expected time of the first bell
            # to infinity so that Wheatley will wait indefinitely for the first bell to ring, and
            # then it will extrapolate from that time
            self._start_time = float('inf')

    # Linear conversions between different time measurements
    def index_to_blow_time(self, row_number: int, place: int) -> float:
        """ Convert a row number and place into a blow_time, taking hanstroke gaps into account. """
        return row_number * self.stage + place + (row_number // 2) * self._handstroke_gap

    def blow_time_to_real_time(self, blow_time: float) -> float:
        """ Convert from blow_time into real_time using the regression line. """
        return self._start_time + self._blow_interval * blow_time

    def index_to_real_time(self, row_number: int, place: int) -> float:
        """
        Convert straight from row number and place into the expected real time according to the
        regression line.
        """
        return self.blow_time_to_real_time(self.index_to_blow_time(row_number, place))

    def real_time_to_blow_time(self, real_time: float) -> float:
        """ Convert backwards from a real time to the corresponding blow time. """
        return (real_time - self._start_time) / self._blow_interval



""" A module for the Rhythm decorator class that adds the ability to wait for human ringers. """

import logging

from typing import Any, Dict, Set

from wheatley.stroke import Stroke, HANDSTROKE, BACKSTROKE
from wheatley.bell import Bell
from .abstract_rhythm import Rhythm



class WaitForUserRhythm(Rhythm):
    """ A decorator class that adds the ability to wait for user-controlled bells to ring. """

    logger_name = "RHYTHM:WaitForUser"
    sleep_time = 0.01

    def __init__(self, rhythm: Rhythm) -> None:
        """
        Initialise a wrapper around another Rhythm class that will decorate that class with the
        ability to wait for other people to ring.
        """
        self._inner_rhythm = rhythm

        self._current_stroke = HANDSTROKE

        self._expected_bells: Dict[Stroke, Set] = {HANDSTROKE: set(), BACKSTROKE: set()}
        self._early_bells: Dict[Stroke, Set] = {HANDSTROKE: set(), BACKSTROKE: set()}

        self.delay = 0.0

        self.logger = logging.getLogger(self.logger_name)

        self._should_return_to_mainloop = False

    def return_to_mainloop(self) -> None:
        """
        When called, this function requires that Wheatley's main thread should return from
        wait_for_bell_time and give control back to the mainloop within a reasonable amount of time
        (i.e. 1/10 of a second is acceptable, but anything approacing a second is not)
        """
        self._should_return_to_mainloop = True
        self._inner_rhythm.return_to_mainloop()

    def wait_for_bell_time(self, current_time: float, bell: Bell, row_number: int, place: int,
                           user_controlled: bool, stroke: Stroke) -> None:
        """ Sleeps the thread until a given Bell should have rung. """
        if stroke != self._current_stroke:
            self.logger.debug(f"Switching to unexpected stroke {stroke}")
            self._current_stroke = stroke

        self._inner_rhythm.wait_for_bell_time(current_time - self.delay, bell, row_number, place,
                                              user_controlled, stroke)
        if user_controlled:
            delay_for_user = 0.0
            self.logger.debug(f"Waiting for {bell}...")
            while bell in self._expected_bells[stroke]:
                self.sleep(self.sleep_time)
                delay_for_user += self.sleep_time
                if self._should_return_to_mainloop:
                    self.logger.debug("Returning to mainloop")
                    break
            self.logger.debug(f"Finished waiting for {bell}")

            if delay_for_user:
                self.delay += delay_for_user
                self.logger.debug(f"Delayed for {delay_for_user:.3f}s")

        # Reset the flag to say that we've returned to the mainloop
        self._should_return_to_mainloop = False

    def expect_bell(self, expected_bell: Bell, row_number: int, place: int, expected_stroke: Stroke) -> None:
        """
        Indicates that a given Bell is expected to be rung at a given row, place and stroke.
        Used by the rhythm so that when that bell is rung later, it can tell where that bell
        _should_ have been in the ringing, and so can use that knowledge to inform the speed of the
        ringing.
        """
        self._inner_rhythm.expect_bell(expected_bell, row_number, place, expected_stroke)

        if expected_stroke != self._current_stroke:
            self._current_stroke = expected_stroke
            self._expected_bells[expected_stroke].clear()
            self._early_bells[expected_stroke.opposite()].clear()

        if expected_bell not in self._early_bells[expected_stroke]:
            self._expected_bells[expected_stroke].add(expected_bell)

    def change_setting(self, key: str, value: Any, real_time: float) -> None:
        # Keep lying to _inner_rhythm about what the current time is
        self._inner_rhythm.change_setting(key, value, real_time - self.delay)

    def on_bell_ring(self, bell: Bell, stroke: Stroke, real_time: float) -> None:
        """
        Called when a bell is rung at a given stroke.  Used as a callback from the Tower class.
        """
        self._inner_rhythm.on_bell_ring(bell, stroke, real_time - self.delay)

        if stroke == self._current_stroke:
            try:
                self._expected_bells[self._current_stroke].remove(bell)
            except KeyError:
                pass
            else:
                self.logger.debug(f"{bell} rung at {stroke}")

            try:
                self._early_bells[self._current_stroke.opposite()].remove(bell)
            except KeyError:
                pass
            else:
                self.logger.debug(f"{bell} reset to {stroke}")
        else:
            self.logger.debug(f"{bell} rung early to {stroke}")
            self._early_bells[self._current_stroke.opposite()].add(bell)

    def initialise_line(self, stage: int, user_controls_treble: bool, start_time: float,
                        number_of_user_controlled_bells: int) -> None:
        """ Allow the Rhythm object to initialise itself when 'Look to' is called. """
        self._expected_bells[HANDSTROKE].clear()
        self._expected_bells[BACKSTROKE].clear()

        self._early_bells[HANDSTROKE].clear()
        self._early_bells[BACKSTROKE].clear()

        self._current_stroke = HANDSTROKE

        # Clear any current waiting loops
        self.sleep(2 * self.sleep_time)

        self._inner_rhythm.initialise_line(stage, user_controls_treble, start_time - self.delay,
                                           number_of_user_controlled_bells)



from unittest import TestCase

from tests.row_generation import as_bells
from wheatley.row_generation import DixonoidsGenerator
from wheatley.stroke import HANDSTROKE, BACKSTROKE


class DixonoidsGeneratorTests(TestCase):

    def test_dixons_treble_leading_handstroke(self):
        generator = DixonoidsGenerator(6, DixonoidsGenerator.DixonsRules)
        # Backstroke previous
        previous_row = as_bells([1, 4, 3, 2, 5, 6])
        new_row = generator._gen_row(previous_row, stroke=HANDSTROKE, index=0)
        self.assertEqual(as_bells([4, 1, 2, 3, 6, 5]), new_row)

    def test_dixons_treble_leading_backstroke(self):
        generator = DixonoidsGenerator(6, DixonoidsGenerator.DixonsRules)
        # Handstroke previous
        previous_row = as_bells([1, 3, 2, 4, 6, 5])
        new_row = generator._gen_row(previous_row, stroke=BACKSTROKE, index=0)
        self.assertEqual(as_bells([1, 3, 4, 2, 5, 6]), new_row)

    def test_dixons_treble_leading_backstroke_bob(self):
        generator = DixonoidsGenerator(6, DixonoidsGenerator.DixonsRules)
        generator.set_bob()
        # Handstroke previous
        previous_row = as_bells([1, 3, 2, 4, 6, 5])
        new_row = generator._gen_row(previous_row, stroke=BACKSTROKE, index=0)
        self.assertEqual(as_bells([1, 2, 3, 4, 5, 6]), new_row)

    def test_dixons_treble_leading_backstroke_single(self):
        generator = DixonoidsGenerator(6, DixonoidsGenerator.DixonsRules)
        generator.set_single()
        # Handstroke previous
        previous_row = as_bells([1, 3, 2, 4, 6, 5])
        new_row = generator._gen_row(previous_row, stroke=BACKSTROKE, index=0)
        self.assertEqual(as_bells([1, 3, 2, 4, 5, 6]), new_row)

    def test_dixons_two_leading_handstroke(self):
        generator = DixonoidsGenerator(6, DixonoidsGenerator.DixonsRules)
        # Backstroke previous
        previous_row = as_bells([2, 4, 1, 3, 5, 6])
        new_row = generator._gen_row(previous_row, stroke=HANDSTROKE, index=0)
        self.assertEqual(as_bells([4, 2, 3, 1, 6, 5]), new_row)

    def test_dixons_two_leading_backstroke(self):
        generator = DixonoidsGenerator(6, DixonoidsGenerator.DixonsRules)
        # Handstroke previous
        previous_row = as_bells([2, 4, 1, 3, 6, 5])
        new_row = generator._gen_row(previous_row, stroke=BACKSTROKE, index=0)
        self.assertEqual(as_bells([2, 1, 4, 3, 5, 6]), new_row)

    def test_dixons_five_leading_handstroke(self):
        generator = DixonoidsGenerator(6, DixonoidsGenerator.DixonsRules)
        # Handstroke previous
        previous_row = as_bells([5, 3, 6, 1, 4, 2])
        new_row = generator._gen_row(previous_row, stroke=HANDSTROKE, index=0)
        self.assertEqual(as_bells([3, 5, 1, 6, 2, 4]), new_row)

    def test_dixons_five_leading_backstroke(self):
        generator = DixonoidsGenerator(6, DixonoidsGenerator.DixonsRules)
        # Handstroke previous
        previous_row = as_bells([5, 6, 3, 4, 1, 2])
        new_row = generator._gen_row(previous_row, stroke=BACKSTROKE, index=0)
        self.assertEqual(as_bells([5, 3, 6, 1, 4, 2]), new_row)



import unittest
from unittest import TestCase

from wheatley.row_generation import MethodPlaceNotationGenerator, generator_from_special_title, PlaceNotationGenerator, \
    PlainHuntGenerator
from wheatley.row_generation.method_place_notation_generator import MethodNotFoundError

plain_bob_minimus = """<?xml version="1.0"?>
<methods xmlns="http://methods.ringing.org/NS/method" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:db="http://methods.ringing.org/NS/database" xmlns:ns_1="http://methods.ringing.org/NS/database" version="0.1" ns_1:page="0" ns_1:pagesize="100" ns_1:rows="1">
  <method xmlns:a="http://methods.ringing.org/NS/method" id="m13199">
    <stage>4</stage>
    <title>Plain Bob Minimus</title>
    <pn>
      <symblock>-14-14</symblock>
      <symblock>12</symblock>
    </pn>
  </method>
</methods>
"""

single_oxford_triples = """<?xml version="1.0"?>
<methods xmlns="http://methods.ringing.org/NS/method" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:db="http://methods.ringing.org/NS/database" xmlns:ns_1="http://methods.ringing.org/NS/database" version="0.1" ns_1:page="0" ns_1:pagesize="100" ns_1:rows="1">
  <method xmlns:a="http://methods.ringing.org/NS/method" id="m14153">
    <stage>7</stage>
    <title>Single Oxford Bob Triples</title>
    <pn>
      <symblock>3</symblock>
      <symblock>1.5.1.7.1.7.1</symblock>
    </pn>
  </method>
</methods>
"""

scientific_triples = """<?xml version="1.0"?>
<methods xmlns="http://methods.ringing.org/NS/method" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:db="http://methods.ringing.org/NS/database" xmlns:ns_1="http://methods.ringing.org/NS/database" version="0.1" ns_1:page="0" ns_1:pagesize="100" ns_1:rows="1">
  <method xmlns:a="http://methods.ringing.org/NS/method" id="m26235">
    <stage>7</stage>
    <title>Scientific Triples</title>
    <pn>
      <block>3.1.7.1.5.1.7.1.7.5.1.7.1.7.1.7.1.7.1.5.1.5.1.7.1.7.1.7.1.7</block>
    </pn>
  </method>
</methods>
"""


class MethodPlaceNotationGeneratorTests(TestCase):
    def test_parse_symblock_even(self):
        method_pn, stage, title = MethodPlaceNotationGenerator._parse_xml(plain_bob_minimus)
        self.assertEqual("&-14-14,&12", method_pn)
        self.assertEqual(4, stage)
        self.assertEqual("Plain Bob Minimus", title)

    def test_parse_symblock_odd(self):
        method_pn, stage, title = MethodPlaceNotationGenerator._parse_xml(single_oxford_triples)
        self.assertEqual("&3,&1.5.1.7.1.7.1", method_pn)
        self.assertEqual(7, stage)
        self.assertEqual("Single Oxford Bob Triples", title)

    def test_parse_block(self):
        method_pn, stage, title = MethodPlaceNotationGenerator._parse_xml(scientific_triples)
        self.assertEqual("3.1.7.1.5.1.7.1.7.5.1.7.1.7.1.7.1.7.1.5.1.5.1.7.1.7.1.7.1.7", method_pn)
        self.assertEqual(7, stage)
        self.assertEqual("Scientific Triples", title)


class SpecialMethodNameTests(TestCase):
    def test_grandsire__all_stages_above_4(self):
        test_cases = [
            ("Grandsire Doubles", 5),
            ("GrandSire DouBles", 5),
            ("grandsire doubles", 5),
            ("GRANDSIRE DOUBLES", 5),
            ("  Grandsire   Doubles  ", 5),
            ("Grandsire 5", 5),
            ("Grandsire Triples", 7),
            ("Grandsire Caters", 9),
            ("Grandsire Cinques", 11),
            ("Grandsire Minor", 6),
            ("Grandsire Major", 8),
            ("Grandsire Royal", 10),
            ("Grandsire Maximus", 12),
        ]

        for (method_title, expected_stage) in test_cases:
            with self.subTest(method_title=method_title, expected_stage=expected_stage):
                generator = generator_from_special_title(method_title)
                self.assertIsInstance(generator, PlaceNotationGenerator)
                self.assertTrue(expected_stage, generator.stage)

    def test_grandsire__in_title__is_none(self):
        test_cases = [
            ("Reverse Grandsire Doubles"),
            ("Double Grandsire Triples"),
            ("Grandsire Five Little Bob Triples")
        ]

        for (method_title) in test_cases:
            with self.subTest(method_title=method_title):
                generator = generator_from_special_title(method_title)
                self.assertIsNone(generator)

    def test_stedman__odd_stages(self):
        test_cases = [
            ("Stedman Doubles", 5),
            ("Stedman Triples", 7),
            ("Stedman Caters", 9),
            ("Stedman Cinques", 11),
        ]

        for (method_title, expected_stage) in test_cases:
            with self.subTest(method_title=method_title, expected_stage=expected_stage):
                generator = generator_from_special_title(method_title)
                self.assertIsInstance(generator, PlaceNotationGenerator)
                self.assertTrue(expected_stage, generator.stage)

    def test_stedman__other_stages__is_none(self):
        test_cases = [
            ("Stedman Singles"),
            ("Stedman Minimus"),
            ("Stedman Minor"),
            ("Stedman Major"),
            ("Stedman Royal"),
            ("Stedman Maximus"),
        ]

        for (method_title) in test_cases:
            with self.subTest(method_title=method_title):
                generator = generator_from_special_title(method_title)
                self.assertIsNone(generator)

    def test_plain_hunt__all_stages(self):
        test_cases = [
            ("Plain Hunt Singles", 3),
            ("Plain Hunt 3", 3),
            ("Plain Hunt on 3", 3),
            ("Plain Hunt Doubles", 5),
            ("Plain Hunt Triples", 7),
            ("Plain Hunt Caters", 9),
            ("Plain Hunt Cinques", 11),
            ("Plain Hunt Minimus", 4),
            ("Plain Hunt Minor", 6),
            ("Plain Hunt Major", 8),
            ("Plain Hunt Royal", 10),
            ("Plain Hunt Maximus", 12),
        ]

        for (method_title, expected_stage) in test_cases:
            with self.subTest(method_title=method_title, expected_stage=expected_stage):
                generator = generator_from_special_title(method_title)
                self.assertIsInstance(generator, PlainHuntGenerator)
                self.assertTrue(expected_stage, generator.stage)

    def test_unknown_stages__throws_MethodNotFoundError(self):
        test_cases = [
            ("Stedman"),
            ("Stedman Not a stage"),
            ("Stedman 0"),
        ]

        for (method_title) in test_cases:
            with self.subTest(method_title=method_title):
                with self.assertRaises(MethodNotFoundError):
                    generator = generator_from_special_title(method_title)


if __name__ == '__main__':
    unittest.main()



""" A module to hold the row generator that generates rows given some place notations. """

from typing import ClassVar, List, Dict, Optional

from wheatley.aliases import CallDef, Row, Places
from wheatley.stroke import Stroke

from .helpers import convert_pn, convert_to_bell_string
from .row_generator import RowGenerator


class PlaceNotationGenerator(RowGenerator):
    """ A row generator to generate rows given a place notation. """

    # Dict Lead Index: String PlaceNotation
    # 0 for end of the lead
    DEFAULT_BOB: ClassVar[CallDef] = CallDef({0: '14'})
    DEFAULT_SINGLE: ClassVar[CallDef] = CallDef({0: '1234'})

    def __init__(self, stage: int, method: str, bob: CallDef = None, single: CallDef = None,
                 start_index: int = 0, start_row: Optional[str] = None) -> None:
        super().__init__(stage, start_row)

        if bob is None:
            bob = PlaceNotationGenerator.DEFAULT_BOB
        if single is None:
            single = PlaceNotationGenerator.DEFAULT_SINGLE

        self.method_pn = convert_pn(method)
        self.lead_len = len(self.method_pn)
        # Store the method place notation as a string for the summary string
        self.method_pn_string = method
        self.start_index = start_index

        def parse_call_dict(unparsed_calls: CallDef) -> Dict[int, List[Places]]:
            """ Parse a dict of type `int => str` to `int => [PlaceNotation]`. """
            parsed_calls = {}

            for i, place_notation_str in unparsed_calls.items():
                # Parse the place notation string into a list of place notations, adjust the
                # call locations by the length of their calls (so that e.g. `0` always refers to
                # the lead end regardless of how long the calls are).
                converted_place_notations = convert_pn(place_notation_str)

                # Add the processed call to the output dictionary
                parsed_calls[(i - 1) % self.lead_len] = converted_place_notations

            return parsed_calls

        self.bobs_pn = parse_call_dict(bob)
        self.singles_pn = parse_call_dict(single)

        self._generating_call_pn: List[Places] = []

    def summary_string(self) -> str:
        """ Returns a short string summarising the RowGenerator. """
        return f"place notation '{self.method_pn_string}'"

    def _gen_row(self, previous_row: Row, stroke: Stroke, index: int) -> Row:
        assert self.lead_len > 0
        lead_index = (index + self.start_index) % self.lead_len

        if self._has_bob and self.bobs_pn.get(lead_index):
            self._generating_call_pn = list(self.bobs_pn[lead_index])
            self.logger.info(f"Bob at index {lead_index}")
            self.reset_calls()
        elif self._has_single and self.singles_pn.get(lead_index):
            self._generating_call_pn = list(self.singles_pn[lead_index])
            self.logger.info(f"Single at index {lead_index}")
            self.reset_calls()

        if self._generating_call_pn:
            place_notation = self._generating_call_pn.pop(0)
        else:
            place_notation = self.method_pn[lead_index]

        return self.permute(previous_row, place_notation)

    def start_stroke(self) -> Stroke:
        return Stroke.from_index(self.start_index)

    @staticmethod
    def grandsire(stage: int, start_row: Optional[str] = None) -> RowGenerator:
        """ Generates Grandsire on a given stage. """
        stage_bell = convert_to_bell_string(stage)

        cross_notation = stage_bell if stage % 2 else '-'

        main_body = ["1" if i % 2 else cross_notation for i in range(2 * stage)]
        main_body[0] = "3"
        notation = ".".join(main_body)

        return PlaceNotationGenerator(stage, notation, bob=CallDef({-1: "3"}), single=CallDef({-1: "3.123"}),
                                      start_row=start_row)

    @staticmethod
    def stedman(stage: int, start_row: Optional[str] = None) -> RowGenerator:
        """ Generates Stedman on a given stage (even bell Stedman will cause an exception). """
        assert stage % 2 == 1

        if stage == 5:
            return PlaceNotationGenerator.stedman_doubles(start_row)

        stage_bell = convert_to_bell_string(stage)
        stage_bell_1 = convert_to_bell_string(stage - 1)
        stage_bell_2 = convert_to_bell_string(stage - 2)

        notation = f"3.1.{stage_bell}.3.1.3.1.3.{stage_bell}.1.3.1"

        return PlaceNotationGenerator(stage, notation, bob=CallDef({3: stage_bell_2, 9: stage_bell_2}),
                                      single=CallDef({
                                          3: f"{stage_bell_2}{stage_bell_1}{stage_bell}",
                                          9: f"{stage_bell_2}{stage_bell_1}{stage_bell}"}),
                                      start_row=start_row)

    @staticmethod
    def stedman_doubles(start_row: Optional[str] = None) -> RowGenerator:
        """ Generates Stedman on a given stage (even bell Stedman will cause an exception). """
        notation = "3.1.5.3.1.3.1.3.5.1.3.1"

        return PlaceNotationGenerator(5, notation, bob=CallDef({}), single=CallDef({6: "345", 12: "145"}),
                                      start_row=start_row)



""" Setup script to install the Ringing Room Bot. """

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("wheatley/version.txt", "r") as fh:
    version = fh.read()

setuptools.setup(
    name="wheatley",
    version=version,

    author="Ben White-Horne",
    author_email="kneasle@gmail.com",

    description="A program that will ring any set of bells in Ringing Room.",
    long_description=long_description,
    long_description_content_type="text/markdown",

    url="https://github.com/Kneasle/wheatley/",

    license="MIT",
    platforms="any",

    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],
    python_requires='>=3.6',
    install_requires=[
        'numpy<=1.19.3; platform_system == "Windows"',
        'numpy; platform_system != "Windows"',
        "requests",
        "python-socketio<5",
        "python-engineio<4",
        "websocket-client"
    ],
    package_data={"wheatley": ["version.txt"]},

    entry_points={'console_scripts': ['wheatley = wheatley.main:main']}
)
